from pathlib import Path


import xarray as xr
import numpy as np
import geopandas as gpd
from shapely.creation import polygons

from emiproc.grids import WGS84_PROJECTED, GeoPandasGrid
from emiproc.inventories import Inventory
from emiproc.profiles.temporal_profiles import (
    CompositeTemporalProfiles,
    DayOfYearProfile,
    HourOfYearProfile,
)


class LPJ_GUESS_Inventory(Inventory):
    """LPJ-GUESS inventory.

    This is an implementation to read netcdf outputs from LPJ-GUESS.

    https://web.nateko.lu.se/lpj-guess/

    """

    def __init__(self, lpj_guess_files: list[Path]):
        """Initialize the inventory.

        Parameters
        ----------
        lpj_guess_files:
            List of paths to the LPJ-GUESS files.
        """
        super().__init__()

        ds = xr.open_mfdataset(lpj_guess_files, combine="by_coords")

        varnames = [
            k
            for k in ds.variables.mapping.keys()
            if k not in ["longitude", "latitude", "time"]
        ]

        varname_to_catsub = {
            "_".join(splitted): ("_".join(splitted[1:]), splitted[0])
            for splitted in map(lambda x: x.split("_"), varnames)
        }

        # Check that the units are all as expceted
        expected = "mg CH4 m-2 d-1"
        for varname in varnames:
            assert (
                ds[varname].units == expected
            ), f"{varname} has units {ds[varname].units} instead of {expected}"

        # Resample the dataset as a dataarray with the categories and substances added as dimensions
        da = ds.to_dataarray(dim="variable")
        # Transform the variable dimension to a multiindex
        da_merged = xr.combine_by_coords(
            [
                da.sel(variable=varname)
                .drop_vars("variable")
                .expand_dims(dim=dict(category=[cat], substance=[sub]))
                for varname, (cat, sub) in varname_to_catsub.items()
            ],
        )
        da_merged["substance"] = da_merged["substance"].astype(str)
        da_merged["category"] = da_merged["category"].astype(str)

        # replace the lat lon by cell
        da_stacked_all = da_merged.stack(cell=("longitude", "latitude"))

        da_stacked = da_stacked_all.drop_vars(["latitude", "longitude"])
        # Use a simple integer index for the cell
        da_stacked["cell"] = np.arange(da_stacked.sizes["cell"])

        # Make the geometry
        lon = da_stacked_all["longitude"].values
        lat = da_stacked_all["latitude"].values

        lon_range = da_merged["longitude"].values
        lat_range = da_merged["latitude"].values

        d_lon = np.diff(lon_range)[0]
        d_lat = np.diff(lat_range)[0]
        # Ensure the spacing is constant
        assert np.allclose(
            np.diff(lon_range), d_lon
        ), "Longitude spacing is not constant"
        assert np.allclose(
            np.diff(lat_range), d_lat
        ), "Latitude spacing is not constant"

        # Reconstruct the grid vertices
        coords = np.array(
            [
                # Bottom left
                [
                    lon - d_lon / 2,
                    lat - d_lat / 2,
                ],
                # Bottom right
                [
                    lon + d_lon / 2,
                    lat - d_lat / 2,
                ],
                # Top right
                [
                    lon + d_lon / 2,
                    lat + d_lat / 2,
                ],
                # Top left
                [
                    lon - d_lon / 2,
                    lat + d_lat / 2,
                ],
            ]
        )

        coords = np.rollaxis(coords, -1, 0)

        # Create the polygons
        geometry = gpd.GeoSeries(polygons(coords), crs="WGS84")
        self.grid = GeoPandasGrid(geometry, shape=(len(lon_range), len(lat_range)))

        # Unit conversion
        # "mg CH4 m-2 d-1" -> "kg / year / cell"
        # Day to Year is already included when we summed
        # kg/mg * m2/cell
        converstion_factor = 1e-6 * geometry.to_crs(WGS84_PROJECTED).area
        da_total = da_stacked.sum(dim="time") * converstion_factor.values

        # Convert to pandas
        df = (
            da_total.stack(catsub=("category", "substance"))
            .drop_vars(["cell"])
            .to_pandas()
        )
        self.gdf = gpd.GeoDataFrame(df, geometry=geometry)
        self.gdfs = {}

        # Gerenerate the profiles
        da_profiles = da_stacked.stack(profiles=("substance", "category", "cell"))
        # Convert to ratios
        da_ratios = (da_profiles / da_profiles.sum(dim="time")).fillna(0.0)
        mask_valid = da_ratios.sum(dim="time") != 0

        # Create the profiles indexes
        da_profiles_indexes = da_ratios.sum(dim="time")
        da_profiles_indexes.values = -np.ones(da_profiles_indexes.shape, dtype=int)

        # Set the values to linear indices
        da_valid_profiles = da_ratios.sel(profiles=mask_valid)
        da_profiles_indexes.loc[mask_valid] = np.arange(
            da_valid_profiles.sizes["profiles"]
        )
        profiles_indexes = da_profiles_indexes.unstack()

        self.set_profiles(
            profiles=CompositeTemporalProfiles.from_ratios(
                da_valid_profiles.values.T, types=[DayOfYearProfile]
            ),
            indexes=profiles_indexes,
        )



"""Command line interface for emiproc.

It is deprecated since v1 but could be used for utilities if needed.
"""
import argparse

if __name__ == '__main__':
    # Create a parser for the command line arguments
    # The help message displays a deprecated notice
    parser = argparse.ArgumentParser(
        description='Deprecated command line use of emiproc. Use emiproc v1.',
    )

    # parse the command line arguments
    args = parser.parse_args()


    parser.print_help()




"""Test that the icon profiles can be saved."""

from emiproc.exports.icon import make_icon_time_profiles, TemporalProfilesTypes
from emiproc.tests_utils.temporal_profiles import oem_const_profile, oem_test_profile
from emiproc.utilities import HOUR_PER_YR


def test_three_profiles():
    dss = make_icon_time_profiles(
        {"cat_const": oem_const_profile, "cat_test": oem_test_profile},
        time_zones=["Europe/Zurich", "UTC", "Asia/Tokyo", "America/Montreal"],
        profiles_type=TemporalProfilesTypes.THREE_CYCLES,
    )
    # Test we have the three data sets
    assert "hourofday" in dss
    assert "dayofweek" in dss
    assert "monthofyear" in dss


def test_hour_of_year():
    dss = make_icon_time_profiles(
        {"cat_const": oem_const_profile, "cat_test": oem_test_profile},
        time_zones=["Europe/Zurich", "UTC", "Asia/Tokyo", "America/Montreal"],
        profiles_type=TemporalProfilesTypes.HOUR_OF_YEAR,
        year=2022,
    )
    assert "hourofyear" in dss



"""Mappings to regroup categories together."""

CH_2_GNFR = {
    # PublicPower
    "GNFR_A": [
        # Waste incinerator we agreed on nov. 22 to put them in the public power category
        "eipkv",  # Punktquellen KVA (Kehrichtverbrennungsanlagen ) == Waste incinerators
    ],
    # Industry
    "GNFR_B": [
        "eipro",  # Flächenquellen Industrie
        "eipwp",  # this is the weitere punktquelle (additional point sources)
        "eipzm",  # Punktquellen Zementwerke (point sources cement plants)
    ],
    # Other stationary combustion (services, residential, agriculture)
    "GNFR_C": [
        "ehare",
        "ehfho",
        "ehfoe",
        "ehgws",
        "eipdh",  # Diensleistungen Holz und Kohle
        "eiprd",  # Dienstleistungen Ol und Gas
        "elfeu",
    ],
    # Fugitives
    "GNFR_D": [
        "eilgk",  # F-Gase: Läden, Gebäude mit Klimaanlagen Refrigeration (F-gases: shops, buildings with air conditioning)
        "eivgn",  # Verluste Gasnetz (losses gas network)
        "evklm",  # F-Gase: Klimaanlagen Motorfahrzeuge (F-gases: air conditioning motor vehicles)
        "evtrk",  # F-Gase: Transporte mit Kühlung  (F-gases: transports with cooling)
    ],
    # Solvents and product use
    "GNFR_E": [
        "eilmi",  # Lösungsmittel Industrie
        "ehlmk",  # Lösungsmittel Konsumprodukte
    ],
    # Road transport
    "GNFR_F": [
        # "evstr_ch4",
        # "evstr_co",
        # "evstr_co2",
        # "evstr_n2o",
        # "evstr_nh3",
        # "evstr_nmvoc",
        # "evstr_nox",
        # "evstr_so2",
        "evstr",  # Strassenverkehr (road transport)
        "evzon",  # Zonenverkehr Kaltstart/Verdampfung (zone traffic, cold start/evaporation)
    ],
    # Shipping
    "GNFR_G": [
        "evsee",
        "evsfa",
        "evsrh",
    ],
    # Aviation
    "GNFR_H": [
        "evfgva",
        "evfzrh",
    ],
    # Offroad mobility
    "GNFR_I": [
        "ehmgh",
        "eibau",
        "eifrz",
        "eilpf",
        "eipis",
        "elfwm",
        "ellwm",
        "evsch",
        "evsra",
    ],
    # Waste
    "GNFR_J": [
        "eidep",  # Deponien
        "eikla",  # Kläranlagen
        "eikmp",  # Kompostierung
        "elabf",  # Abfallverbrennung Land- und Forstwirtschaft
        "elver",  # Vergärung
    ],
    # AgriLivestock
    "GNFR_K": [
        "elapp",
        "elsto",
    ],
    #  AgriOther
    "GNFR_L": [
        "elfer",
    ],
    # Others
    "GNFR_R": [
        "ehhab",  # Haushalte andere brande
        "ehhaf",  # Haushalte andere feuerwerk
        "ehhan",  # Haushalte andere private
        "enwal",  # Emissionen aus Waldern (emissions from forests)
    ],
}


TNO_2_GNFR = {
    # PublicPower
    "GNFR_A": [
        "A",
    ],
    # Industry
    "GNFR_B": [
        "B",
    ],
    # Other stationary combustion (services, residential, agriculture)
    "GNFR_C": [
        "C",
    ],
    # Fugitives
    "GNFR_D": [
        "D",
    ],
    # Solvents and product use
    "GNFR_E": [
        "E",
    ],
    # Road transport
    "GNFR_F": [
        "F1",
        "F2",
        "F3",
        "F4",
    ],
    # Shipping
    "GNFR_G": [
        "G",
    ],
    # Aviation
    "GNFR_H": [
        "H",
    ],
    # Offroad mobility
    "GNFR_I": [
        "I",
    ],
    # Waste
    "GNFR_J": [
        "J",
    ],
    # AgriLivestock
    "GNFR_K": [
        "K",
    ],
    # AgriOther
    "GNFR_L": [
        "L",
    ],
    # Others
    "GNFR_R": [],
}



from datetime import datetime

NetcdfAttributes = dict[str, str]


def nc_cf_attributes(
    author: str = "emiproc",
    contact: str = "https://emiproc.readthedocs.io/en/master/support.html#contact",
    title: str = "emiproc generated file",
    source: str = "",
    comment: str = "",
    institution: str = "Empa, Swiss Federal Laboratories for Materials Science and Technology",
    history: str = "",
    references: str = "Produced by emiproc.",
    additional_attributes: NetcdfAttributes = {},
) -> NetcdfAttributes:
    """Create attributes for a nc file based on cf conventions.


    Most of the following instructions at
    https://cfconventions.org/Data/cf-conventions/cf-conventions-1.10/cf-conventions.html#_attributes

    Some additional fields to the cf-conventions (author, contact)
    are added for simplifying data sharing.

    :param author: The author of the file.
    :param contact: The contact address (email) of a person that can be contacted
        if one has questions about the file.
    :param title:
        A succinct description of what is in the dataset.
    :param institution:
        Specifies where the original data was produced.
    :param source:
        The method of production of the original data. If it was model-generated, source should name the model and its version, as specifically as could be useful. If it is observational, source should characterize it (e.g., "surface observation" or "radiosonde").
    :param history:
        Provides an audit trail for modifications to the original data.
        Well-behaved generic netCDF filters will automatically append their name and the parameters with which they were invoked to the global history attribute of an input netCDF file
        We recommend that each line begin with a timestamp indicating the date and time of day that the program was executed.
    :param references:
        Published or web-based references that describe the data or methods used to produce it.
    :param comment:
        Miscellaneous information about the data or methods used to produce it.
    :param additional_attributes:
        Any attribute you want to add to the netcdf file produced.

    """
    dt = datetime.now().strftime("%Y-%m-%d %H:%M:%S")
    return {
        "Conventions": "CF-1.10",
        "title": title,
        "comment": comment,
        "source": source,
        "history": dt + ": created by emiproc ;\n" + history,
        "references": references,
        "institution": institution,
        "author": author,
        "contact": contact,
        "creation_time": dt,
        **additional_attributes,
    }


DEFAULT_NC_ATTRIBUTES = nc_cf_attributes(
    author="emiproc",
    contact="https://emiproc.readthedocs.io/en/latest/support.html#contact",
    title="emiproc generated file",
    source="unspecified",
    comment="This file was automatically generated by emiproc",
)



"""Show some basic features of the temporal profiles and plot."""

# %%
from datetime import datetime, timedelta
from os import PathLike
import xarray as xr
from pathlib import Path
import numpy as np
import pytest
import geopandas as gpd
from shapely.geometry import Polygon
from emiproc.exports.icon import TemporalProfilesTypes, make_icon_time_profiles

from emiproc.inventories import Inventory
from emiproc.tests_utils.test_inventories import inv, inv_with_pnt_sources
from emiproc.profiles.temporal_profiles import (
    TemporalProfile,
    DailyProfile,
    WeeklyProfile,
    MounthsProfile,
    create_scaling_factors_time_serie,
    from_csv,
    from_yaml,
)
from emiproc import FILES_DIR

# %% test

DailyProfile()
# %% Create test profiles
test_profiles = [
    DailyProfile(),
    WeeklyProfile(),
    MounthsProfile(),
    TemporalProfile(
        size=3,
        ratios=np.array([0.3, 0.2, 0.5]),
    ),
]

# %%
ratios = create_scaling_factors_time_serie(
    start_time="2020-01-01",
    end_time="2020-01-31",
    profiles=[DailyProfile(), WeeklyProfile(), MounthsProfile()],
)
ratios

# %%
profiles_dir = FILES_DIR / "profiles" / "copernicus"
profiles_month_in_year = from_csv(
    profiles_dir / "timeprofiles-month_in_year.csv"
)
profiles_day_in_week = from_csv(
    profiles_dir / "timeprofiles-day_in_week.csv"
)
profiles_hour_in_day = from_csv(
    profiles_dir / "timeprofiles-hour_in_day.csv"
)
# %% plot the profiles
import matplotlib.pyplot as plt
import pandas as pd

fig, ax = plt.subplots(1, 1, figsize=(10, 5))
for name, profile in profiles_month_in_year.items():
    ax.plot(profile.ratios, label=name)
ax.legend()
# %%
tss = {}
csv_profiles = {}
for categorie in profiles_month_in_year.keys():
    csv_profiles[categorie] = [
        profiles_month_in_year[categorie],
        profiles_day_in_week[categorie],
        profiles_hour_in_day[categorie],
    ]

    tss[categorie] = create_scaling_factors_time_serie(
        start_time="2020-01-01",
        end_time="2020-02-28",
        profiles=csv_profiles[categorie],
    )


fig, ax = plt.subplots(1, 1, figsize=(10, 5))
for name, ts in tss.items():
    ax.plot(ts, label=name)
ax.legend()
# %%
yaml_dir = Path(FILES_DIR / "profiles" / "yamls")

yaml_profiles = {}
for yml_file in yaml_dir.glob("*.yaml"):
    yaml_profiles[yml_file.stem] = from_yaml(yml_file)

# %%
tss = {}
for categorie in yaml_profiles.keys():
    tss[categorie] = create_scaling_factors_time_serie(
        start_time="2020-01-01",
        end_time="2020-02-28",
        profiles=yaml_profiles[categorie],
    )

fig, ax = plt.subplots(1, 1, figsize=(10, 5))
for name, ts in tss.items():
    ax.plot(ts, label=name)
ax.legend()
# %%


dss = make_icon_time_profiles(
    csv_profiles,
    {"DE": 0, "FR": 1, "USA": -5},
    #{"DE": 0, "FR": 1},
    TemporalProfilesTypes.HOUR_OF_YEAR,
    2020,
    out_dir="./test_icon_oem_profiles",
)
dss['hourofyear']
# %%

# plot each of the countries
fig, ax = plt.subplots(1, 1, figsize=(10, 5))
for country in dss["hourofyear"]["country"].values:
    ax.plot(
        dss["hourofyear"]["Road_Transport"].sel(country=country),
        label=country,
        alpha=0.5,
    )
ax.legend()

# %%



"""Temporal profiles."""

from __future__ import annotations

import logging
from dataclasses import dataclass, field
from datetime import datetime, timedelta
from enum import Enum, auto
from os import PathLike
from pathlib import Path
from typing import Any, Type, Union

import numpy as np
import pandas as pd
import xarray as xr
import yaml

import emiproc
from emiproc.profiles.utils import (
    get_profiles_indexes,
    load_country_tz,
    merge_indexes,
    ratios_dataarray_to_profiles,
    ratios_to_factors,
    read_profile_csv,
    read_profile_file,
)

logger = logging.getLogger(__name__)

# Constants
N_HOUR_DAY = 24
N_DAY_WEEK = 7
N_MONTH_YEAR = 12
N_DAY_YEAR = 365
N_DAY_LEAPYEAR = 366
N_HOUR_WEEK = N_HOUR_DAY * N_DAY_WEEK
N_HOUR_YEAR = N_DAY_YEAR * N_HOUR_DAY
N_HOUR_LEAPYEAR = N_DAY_LEAPYEAR * N_HOUR_DAY


class SpecificDay(Enum):
    """An enum to define specific day applied of a profile."""

    # Make it automatically assign the value to the name
    def _generate_next_value_(name: str, start, count, last_values):
        return name.lower()

    MONDAY = auto()
    TUESDAY = auto()
    WEDNESDAY = auto()
    THURSDAY = auto()
    FRIDAY = auto()
    SATURDAY = auto()
    SUNDAY = auto()

    # One of the 5 first days
    WEEKDAY = auto()

    # One of the 2 last days
    WEEKEND = auto()

    @classmethod
    def from_day_number(cls, day_number: int) -> SpecificDay:
        """Return the day corresponding to the day number.

        Day start at 0 for Monday and ends at 6 for Sunday.
        """
        if not isinstance(day_number, int):
            raise TypeError(f"{day_number=} must be an int.")

        dict_day = {
            0: cls.MONDAY,
            1: cls.TUESDAY,
            2: cls.WEDNESDAY,
            3: cls.THURSDAY,
            4: cls.FRIDAY,
            5: cls.SATURDAY,
            6: cls.SUNDAY,
        }

        if day_number not in dict_day:
            raise ValueError(f"{day_number=} is not a valid day number.")

        return dict_day[day_number]

    def __lt__(self, other: SpecificDay) -> bool:
        """Compare the days by the order of the week."""
        if not isinstance(other, SpecificDay):
            raise TypeError(f"{other=} must be a {SpecificDay}.")
        return self.value < other.value


def days_of_specific_day(specific_day: SpecificDay) -> list[SpecificDay]:
    int_days = get_days_as_ints(specific_day)
    return [SpecificDay.from_day_number(i) for i in int_days]


def get_days_as_ints(specific_day: SpecificDay) -> list[int]:
    """Return the days corresponding for a specific day."""

    if not isinstance(specific_day, SpecificDay):
        raise TypeError(f"{specific_day=} must be a {SpecificDay}.")

    if specific_day == SpecificDay.MONDAY:
        return [0]
    elif specific_day == SpecificDay.TUESDAY:
        return [1]
    elif specific_day == SpecificDay.WEDNESDAY:
        return [2]
    elif specific_day == SpecificDay.THURSDAY:
        return [3]
    elif specific_day == SpecificDay.FRIDAY:
        return [4]
    elif specific_day == SpecificDay.SATURDAY:
        return [5]
    elif specific_day == SpecificDay.SUNDAY:
        return [6]
    elif specific_day == SpecificDay.WEEKDAY:
        return [0, 1, 2, 3, 4]
    elif specific_day == SpecificDay.WEEKEND:
        return [5, 6]
    else:
        raise NotImplementedError(
            f"{specific_day=} is implemented in  {get_days_as_ints}"
        )


COUNTRY_TZ_DF = None


def get_emep_shift(country_code_iso3: str) -> int:
    """Retunr the time shift form the country code of emep."""
    global COUNTRY_TZ_DF
    # if COUNTRY_TZ_DF is None:
    COUNTRY_TZ_DF = load_country_tz()

    return COUNTRY_TZ_DF.loc[country_code_iso3, "timezone"]


def concatenate_time_profiles(profiles: list[AnyTimeProfile]) -> AnyTimeProfile:
    """Concatenate the time profiles in the list."""
    if not isinstance(profiles, list):
        raise TypeError(f"{profiles=} must be a list.")

    if not profiles:
        raise ValueError(f"{profiles=} must not be empty.")

    if len(profiles) == 1:
        return profiles[0]

    # Get the type of the profiles
    profile_type = type(profiles[0])

    # Check that all the profiles are of the same type
    if not all(isinstance(p, profile_type) for p in profiles):
        raise ValueError(
            f"{profiles=} must all be of the same type, got {profile_type=}."
        )

    # Concatenate the ratios
    ratios = np.concatenate([p.ratios for p in profiles])

    return profile_type(ratios)


@dataclass(eq=False)
class TemporalProfile:
    """Temporal profile.

    Temporal profile defines how the emission is distributed over time.
    """

    ratios: np.ndarray | None = None
    size: int = 1

    def __post_init__(self) -> None:
        # Check the size is an int
        if not isinstance(self.size, int):
            raise TypeError(f"{self.size=} must be an int.")
        if self.size < 1:
            raise ValueError(f"{self.size=} must be positive.")

        if self.ratios is None:
            self.ratios = np.ones((1, self.size)) / self.size
        else:
            if isinstance(self.ratios, list):
                self.ratios = np.array(self.ratios)
            if len(self.ratios.shape) == 1:
                self.ratios = self.ratios.reshape((1, -1))
            # Make sure the size is a int and the array has the correct size
            if self.size != self.ratios.shape[1]:
                raise ValueError(
                    f"{self.ratios.shape[1]=} does not match profile's {self.size=}."
                )

            # Make sure the ratios sum up to 1
            if not np.all(np.isclose(self.ratios.sum(axis=1), 1.0)):
                raise ValueError(f"{self.ratios.sum(axis=1)=} are not all 1.")

    @property
    def n_profiles(self) -> int:
        return self.ratios.shape[0]

    def __getitem__(self, key: int) -> TemporalProfile:
        """Return the profile at the given index."""
        return self.__class__(self.ratios[key])

    def __len__(self) -> int:
        return self.n_profiles

    def __iter__(self):
        for i in range(self.n_profiles):
            yield self[i]

    # Defie the equality of the profiles
    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, type(self)):
            raise TypeError(f"{other=} must be a {type(self)}.")
        if self.n_profiles != other.n_profiles:
            return False
        return (self.ratios == other.ratios).all()

    # Defined greater smaller by the size attribute
    def __lt__(self, other: AnyTimeProfile) -> bool:
        # All subclasses can be compared
        if not isinstance(other, TemporalProfile):
            raise TypeError(f"{other=} must be a {TemporalProfile}.")
        return self.size < other.size


@dataclass(eq=False)
class DailyProfile(TemporalProfile):
    """Daily profile.

    Daily profile defines how the emission is distributed over the day.
    The profile starts at 00:00.
    """

    size: int = field(default=N_HOUR_DAY, init=False)


@dataclass(eq=False)
class SpecificDayProfile(DailyProfile):
    """Same as DailyProfile but with a specific day of the week."""

    specific_day: SpecificDay | None = None

    def __post_init__(self) -> None:
        super().__post_init__()

        if self.specific_day is None:
            raise ValueError(f"{self.specific_day=} must be defined.")

    def __getitem__(self, key: int) -> SpecificDayProfile:
        """Return the profile at the given index."""
        return SpecificDayProfile(self.ratios[key], specific_day=self.specific_day)

    # Defie the equality of the profiles
    def __eq__(self, other: Any) -> bool:
        super_result = super().__eq__(other)
        if isinstance(other, SpecificDayProfile):
            return super_result and (self.specific_day == other.specific_day)
        else:
            return super_result

    # Defined greater smaller by the size attribute
    def __lt__(self, other: AnyTimeProfile) -> bool:
        # sort by the specific day
        if isinstance(other, SpecificDayProfile):
            return self.specific_day < other.specific_day
        return super().__lt__(other)


@dataclass(eq=False)
class WeeklyProfile(TemporalProfile):
    """Weekly profile.

    Weekly profile defines how the emission is distributed over the week.
    The profile starts on Monday.
    """

    size: int = N_DAY_WEEK


@dataclass(eq=False)
class MounthsProfile(TemporalProfile):
    """Yearly profile.

    Yearly profile defines how the emission is distributed over the year using months.
    """

    size: int = N_MONTH_YEAR


@dataclass(eq=False)
class HourOfWeekProfile(TemporalProfile):
    """Hour of week profile.

    Hour of week profile defines how the emission is distributed over the week using hours.
    This is useful if you want to account for different daily patterns over the
    days of the week (usually for the weekend)

    The profile starts on Monday at 00:00.
    """

    size: int = N_HOUR_WEEK


@dataclass(eq=False)
class HourOfYearProfile(TemporalProfile):
    """Hour of year profile.

    Hour of year profile defines how the emission is distributed over the year using hours.
    """

    size: int = N_HOUR_YEAR


@dataclass(eq=False)
class Hour3OfDay(TemporalProfile):
    """Each 3 hour profile.

    Groups of 3 hours. (0-3, 3-6, 6-9, 9-12, 12-15, 15-18, 18-21, 21-24)
    """

    size: int = int(N_HOUR_DAY / 3)


@dataclass(eq=False)
class Hour3OfDayPerMonth(TemporalProfile):
    """Each 3 hour profile given for each mounth.

    Hour of year profile defines how the emission is distributed over the year using hours.
    """

    size: int = int(N_HOUR_DAY / 3) * N_MONTH_YEAR


@dataclass(eq=False)
class HourOfLeapYearProfile(TemporalProfile):
    """Hour of leap year profile.

    Hour of leap year profile defines how the emission is distributed over the year using hours.
    """

    size: int = N_HOUR_LEAPYEAR


@dataclass(eq=False)
class DayOfYearProfile(TemporalProfile):
    """Day of year profile.

    Day of year profile defines how the emission is distributed over the year using days.
    """

    size: int = N_DAY_YEAR


@dataclass(eq=False)
class DayOfLeapYearProfile(TemporalProfile):
    """Day of leap year profile.

    Day of leap year profile defines how the emission is distributed over the year using days.
    """

    size: int = N_DAY_LEAPYEAR


AnyTimeProfile = Union[
    DailyProfile,
    SpecificDayProfile,
    WeeklyProfile,
    MounthsProfile,
    HourOfYearProfile,
    HourOfLeapYearProfile,
]

# Maps temporal profiles to their corrected version
leap_year_corrected: dict[TemporalProfile, TemporalProfile] = {
    HourOfYearProfile: HourOfLeapYearProfile,
    DayOfYearProfile: DayOfLeapYearProfile,
}


def get_leap_year_or_normal(
    profile_type: type[TemporalProfile], year: int
) -> type[TemporalProfile]:
    """Return the profile type for the given year."""

    if year % 4 == 0:
        return leap_year_corrected.get(profile_type, profile_type)
    return profile_type


class AnyProfiles:
    """SAme a as temporal profiles, but can store any kind of proifles."""

    _profiles: list[AnyTimeProfile]

    def __init__(self, profiles: list[AnyTimeProfile] = None) -> None:
        if profiles is None:
            profiles = []
        self._profiles = profiles

    def __getitem__(self, key: int) -> AnyTimeProfile:
        for profile in self._profiles:
            if key < profile.n_profiles:
                return profile[key]
            key -= profile.n_profiles
        raise IndexError(f"{key=} is out of range.")

    def append(self, profile: AnyTimeProfile) -> None:
        self._profiles.append(profile)

    def __len__(self) -> int:
        return sum(len(profile) for profile in self._profiles)

    # Iteration must smoothly iterate over the profiles
    def __iter__(self):
        for profiles in self._profiles:
            yield from profiles

    @property
    def n_profiles(self) -> int:
        """Return the number of profiles."""
        return len(self)


class CompositeTemporalProfiles:
    """A helper class to handle mixtures of temporal profiles.

    Acts similar to a TemporalProfile

    Stores a dict for each type of profile,
    """

    _profiles: dict[
        type[AnyTimeProfile] | tuple[type[SpecificDayProfile], SpecificDay],
        AnyTimeProfile,
    ]
    # Store for each type, the indexes of the profiles
    _indexes: dict[
        type[AnyTimeProfile] | tuple[type[SpecificDayProfile], SpecificDay] | None,
        np.ndarray[int],
    ]

    def __init__(self, profiles: list[list[AnyTimeProfile]] = []) -> None:
        n = len(profiles)
        self._profiles = {}
        profiles_lists = {}
        self._indexes = {}
        # Get the unique types of profiles
        types = set(
            t if (t := type(p)) != SpecificDayProfile else (t, p.specific_day)
            for profiles_list in profiles
            for p in profiles_list
        )

        if len(types) == 0:
            # Empty profiles
            # only empty lists given
            self._indexes[None] = np.full(n, fill_value=-1, dtype=int)
            return

        # Allocate arrays
        for profile_type in types:
            if not isinstance(profile_type, tuple) and not issubclass(
                profile_type, TemporalProfile
            ):
                raise TypeError(
                    f"Profiles must be subclass of {TemporalProfile}. Not"
                    f" {profile_type=}"
                )
            profiles_lists[profile_type] = []
            self._indexes[profile_type] = np.full(n, fill_value=-1, dtype=int)
        # Construct the list and indexes based on the input
        for i, profiles_list in enumerate(profiles):
            if not isinstance(profiles_list, list):
                raise TypeError(
                    f"{profiles_list=} must be a list of {TemporalProfile}."
                )
            for profile in profiles_list:
                if profile.n_profiles != 1:
                    raise ValueError(
                        "Can only build CompositeTemporalProfiles from profiles with"
                        f" {profile.n_profiles=}, got {profile=}."
                    )
                p_type = type(profile)
                if p_type == SpecificDayProfile:
                    p_type = (p_type, profile.specific_day)
                list_this_type = profiles_lists[p_type]
                if self._indexes[p_type][i] != -1:
                    raise ValueError(
                        f"Cannot add {profile=} to {self=} as it was already added."
                    )
                self._indexes[p_type][i] = len(list_this_type)
                list_this_type.append(profile)
        # Convert the lists to arrays
        for profile_type, profiles_list in profiles_lists.items():
            ratios = np.concatenate([p.ratios for p in profiles_list])
            if isinstance(profile_type, tuple):
                profile = profile_type[0](
                    ratios=ratios,
                    specific_day=profile_type[1],
                )
            else:
                profile = profile_type(ratios=ratios)
            self._profiles[profile_type] = profile

    def __repr__(self) -> str:
        return f"CompositeProfiles({len(self)} profiles from {[t.__name__ for t in self.types]})"

    def __len__(self) -> int:
        indexes_len = [len(indexes) for indexes in self._indexes.values()]
        if not indexes_len:
            return 0
        # Make sure they are all equal
        if len(set(indexes_len)) != 1:
            raise ValueError(
                f"{self=} has different lengths of indexes for each profile type."
                f" {indexes_len=}"
            )
        return indexes_len[0]

    @property
    def n_profiles(self) -> int:
        """Return the number of profiles."""
        return len(self)

    def __getitem__(self, key: int) -> list[AnyTimeProfile]:
        return [
            self._profiles[p_type][index]
            for p_type, indexes in self._indexes.items()
            if p_type is not None and (index := indexes[key]) != -1
        ]

    def __setitem__(self, key: int, value: list[AnyTimeProfile]) -> None:
        self._array[key] = np.array(value, dtype=object)

    @property
    def types(self) -> list[AnyTimeProfile]:
        """Return the types of the profiles."""
        return list(set(self._profiles.keys()))

    @property
    def ratios(self) -> np.ndarray:
        """Return ratios of composite profiles.

        Idea is that we concatenate the ratio of each profile.
        nan values can be used when a profile is not defined for a given index.
        """
        return np.stack(
            [
                np.concatenate(
                    [
                        (
                            self._profiles[pt][index].ratios.reshape(-1)
                            if (index := self._indexes[pt][i]) != -1
                            else np.full(pt.size, np.nan).reshape(-1)
                        )
                        for pt in self.types
                    ]
                )
                for i in range(len(self))
            ],
            # axis=1,
        )

    @property
    def scaling_factors(self) -> np.ndarray:
        """Return the scaling factors of the profiles."""
        return np.stack(
            [
                np.concatenate(
                    [
                        (
                            self._profiles[pt][index].ratios.reshape(-1)
                            * self._profiles[pt][index].size
                            if (index := self._indexes[pt][i]) != -1
                            else np.ones(pt.size).reshape(-1)
                        )
                        for pt in self.types
                    ]
                )
                for i in range(len(self))
            ],
            # axis=1,
        )

    @classmethod
    def from_ratios(
        cls, ratios: np.ndarray, types: list[type], rescale: bool = False
    ) -> CompositeTemporalProfiles:
        """Create a composite profile, directly from the ratios.

        :arg ratios: The ratios of the profiles.
        :arg types: The types of the profiles, as a list of Temporal profiles types.
        :arg rescale: If True, the ratios will be rescaled to sum up to 1.

        """
        for t in types:
            # Check that the type is a subtype of TemporalProfile
            if not issubclass(t, TemporalProfile):
                raise TypeError(f"{t=} must be a {TemporalProfile}.")
        splitters = np.cumsum([0] + [t.size for t in types])
        logger.debug(f"{splitters=}")
        # Create the empty profiles
        profiles = [
            [
                t((r / r.sum(axis=0)) if rescale else r)
                for i, t in enumerate(types)
                if not np.any(
                    np.isnan(r := profile_ratios[splitters[i] : splitters[i + 1]])
                )
            ]
            for profile_ratios in ratios
        ]
        return cls(profiles)

    def __eq__(self, __value: object) -> bool:
        if not isinstance(__value, CompositeTemporalProfiles):
            raise TypeError(f"{__value=} must be a {CompositeTemporalProfiles}.")
        if len(self) != len(__value):
            return False
        return (self.ratios == __value.ratios).all()

    @classmethod
    def join(cls, *profiles: CompositeTemporalProfiles) -> CompositeTemporalProfiles:
        """Join multiple composite profiles."""
        # Get the types of profiles
        _profiles = {}
        types = set(sum((p.types for p in profiles), []))
        profile_lenghts = [len(p) for p in profiles]
        total_len = sum(profile_lenghts)
        _indexes = {t: np.full(total_len, fill_value=-1, dtype=int) for t in types}
        for t in types:
            _this_type_profiles = [p._profiles[t] for p in profiles if t in p.types]
            _this_type_n_profiles = [
                len(p._profiles[t]) if t in p.types else 0 for p in profiles
            ]
            _profiles[t] = concatenate_time_profiles(_this_type_profiles)

            # offset in the indexes indexes
            curr_index = 0
            # offset in the profile indexes
            curr_profile = 0
            for i, n in enumerate(_this_type_n_profiles):
                if n == 0:
                    curr_index += profile_lenghts[i]
                    continue
                indexes = profiles[i]._indexes[t].copy()
                mask_invalid = indexes == -1
                indexes[~mask_invalid] += curr_profile
                _indexes[t][curr_index : curr_index + len(indexes)] = indexes
                curr_index += profile_lenghts[i]
                curr_profile += n

        # Get the indexes
        obj = cls([])
        obj._profiles = _profiles
        obj._indexes = _indexes

        return obj

    def copy(self) -> CompositeTemporalProfiles:
        """Return a copy of the object."""
        return CompositeTemporalProfiles.join(self)

    # define the addition to be the same as if this was a list
    def __add__(self, other: CompositeTemporalProfiles) -> CompositeTemporalProfiles:
        return self.join(self, other)

    def __radd__(self, other: CompositeTemporalProfiles) -> CompositeTemporalProfiles:
        return self.join(other, self)

    def append(self, profiles_list: list[AnyTimeProfile]) -> None:
        """Append a profile list to this."""
        new_len = len(self) + 1
        original_types = self.types
        # expend all the indexes list
        for t in self._indexes.keys():
            self._indexes[t] = np.concatenate(
                (
                    self._indexes[t],
                    np.array([-1], dtype=int),
                )
            )

        for p in profiles_list:
            t = type(p)
            if t == SpecificDayProfile:
                t = (SpecificDayProfile, p.specific_day)
            if t not in self._indexes.keys():
                if isinstance(t, tuple):
                    self._profiles[t] = t[0](ratios=p.ratios, specific_day=t[1])
                else:
                    self._profiles[t] = t(ratios=p.ratios)
                self._indexes[t] = np.full(new_len, fill_value=-1, dtype=int)
                self._indexes[t][-1] = 0
            else:
                self._indexes[t][-1] = len(self._profiles[t])
                self._profiles[t].ratios = np.concatenate(
                    (self._profiles[t].ratios, p.ratios)
                )

    @property
    def size(self) -> int:
        """Return the size of the profiles."""
        return sum(p.size for p in self._profiles.values())

    def broadcast(self, types: list[TemporalProfile]) -> CompositeTemporalProfiles:
        """Create a new composite profile with the given types.

        The non specified profiles will be set to constant profiles.
        """
        all_ratios = []
        for t in types:
            # Get constant ratios
            composite_ratios = np.ones((len(self), t.size)) / t.size
            if t in self.types:
                ratios = self._profiles[t].ratios
                # Need to scale with the indexes
                indexes = self._indexes[t]
                # Fill the ratios with the existing profiles
                mask_valid = indexes != -1
                composite_ratios[mask_valid, :] = ratios[indexes[mask_valid], :]

            all_ratios.append(composite_ratios)

        return CompositeTemporalProfiles.from_ratios(
            np.concatenate(all_ratios, axis=1), types
        )


def make_composite_profiles(
    profiles: AnyProfiles,
    indexes: xr.DataArray,
) -> tuple[CompositeTemporalProfiles, xr.DataArray]:
    """Create a composite temporal profiles from a list of profiles and indexes.

    :arg profiles: The profiles to use.
    :arg indexes: The indexes to use.
        The indexes must have a dim called "profile" with the name of the profile type.

    """

    if not isinstance(profiles, AnyProfiles):
        raise TypeError(f"{profiles=} must be an {AnyProfiles}.")

    logger.debug(f"making composite profiles from {profiles=}, {indexes=}")

    if "profile" not in indexes.dims:
        raise ValueError(f"{indexes=} must have a dim called 'profile'.")
    # If size of the profiles is 1, then we can simply return the profiles
    if indexes.profile.size == 1:
        # It is only one type of profile
        return CompositeTemporalProfiles([[p] for p in profiles]), indexes.squeeze(
            "profile"
        )

    # Stack the arrays to keep only the profiles dimension and the new stacked dim
    dims = list(indexes.dims)
    dims.remove("profile")
    stacked = indexes.stack(z=dims)

    str_array = np.array(
        [
            str(array.values.reshape(-1))
            for lab, array in stacked.groupby(group="z", squeeze=False)
        ]
    )
    logger.debug(f"{str_array=}")
    u, inv = np.unique(str_array, return_inverse=True)

    extracted_profiles = [
        [
            profiles[i]
            # Unpakc the profiles from the str
            for i in np.fromstring(array_str[1:-1], sep=" ", dtype=int)
            if i != -1
        ]
        # Loop over each unique profile found
        for array_str in u
    ]
    logger.debug(f"{extracted_profiles=}")
    new_indexes = xr.DataArray(inv, dims=["z"], coords={"z": stacked.z})

    # Remove the z dimension from the profiles
    out_indexes = new_indexes.unstack("z")

    return CompositeTemporalProfiles(extracted_profiles), out_indexes


def get_index_in_profile(
    profile: Type[TemporalProfile], time_range: pd.DatetimeIndex
) -> pd.Series:
    """Get the index in the profile for each time in the time range.

    :param profile: the profile to use
    :param time_range: the time range to use
    :return: the index in the profile for each time in the time range
    """

    if profile == MounthsProfile:
        indexes = time_range.month - 1
    elif profile in [DayOfYearProfile, DayOfLeapYearProfile]:
        indexes = time_range.day_of_year - 1
    elif profile == DailyProfile:
        indexes = time_range.hour
    elif profile == WeeklyProfile:
        indexes = time_range.day_of_week
    elif profile in [HourOfYearProfile, HourOfLeapYearProfile]:
        indexes = time_range.hour + (time_range.day_of_year - 1) * 24
    elif profile == Hour3OfDayPerMonth:
        indexes = (time_range.hour // 3) + (time_range.month - 1) * 8
    else:
        raise ValueError(f"Profile type {profile} not recognized")

    assert indexes.min() >= 0, f"{profile=}, {time_range=}"
    assert indexes.max() < profile.size, f"{profile=}, {time_range=}"

    return indexes


def get_profile_da(
    profile: TemporalProfile, year: int, use_offset: bool = True
) -> xr.DataArray:
    """Return the profile as a data array.

    The index of the data array is exact timestamp at the middle of the period.
    """
    daterange_kwargs = {
        "start": f"{year}-01-01",
        "end": f"{year+1}-01-01",
        "inclusive": "both",
    }

    # The following will create correct timestamps at which the profile is true
    # An offset is also given, which is half the period

    if isinstance(profile, (DailyProfile, HourOfYearProfile, HourOfLeapYearProfile)):
        ts = pd.date_range(**daterange_kwargs, freq="h")
        offset = pd.Timedelta("30m")
    elif isinstance(profile, (WeeklyProfile, DayOfLeapYearProfile, DayOfYearProfile)):
        ts = pd.date_range(**daterange_kwargs, freq="d")
        offset = pd.Timedelta("12h")
    elif isinstance(profile, MounthsProfile):
        ts = pd.date_range(**daterange_kwargs, freq="MS")
        offset = pd.Timedelta("15d")
    elif isinstance(profile, Hour3OfDayPerMonth):
        ts = pd.date_range(**daterange_kwargs, freq="3h")
        offset = pd.Timedelta("1h30m")
    else:
        raise NotImplementedError(
            f"{type(profile)=} not implemented in `get_profile_da`."
        )

    # Add a first value at the begginning, such that we cover the whole year
    ts = pd.DatetimeIndex([ts[0] - 2 * offset] + list(ts))

    # Correct for non cyclic profiles (day of specific year)
    if isinstance(
        profile,
        (
            HourOfYearProfile,
            HourOfLeapYearProfile,
            DayOfLeapYearProfile,
            DayOfYearProfile,
        ),
    ):
        ts = ts[1:-1]

    da = xr.DataArray(
        profile.ratios[:, get_index_in_profile(type(profile), ts)],
        dims=["profile", "datetime"],
        coords={
            "profile": range(profile.n_profiles),
            "datetime": ts + offset if use_offset else ts,
        },
    )

    return da


def interpolate_profiles_hour_of_year(
    profiles: CompositeTemporalProfiles,
    year: int,
    interpolation_method: str = "linear",
    return_profiles: bool = False,
) -> (
    CompositeTemporalProfiles
    | xr.DataArray
    | tuple[CompositeTemporalProfiles | xr.DataArray, xr.DataArray]
):
    """Interpolate the profiles to create a hour of year profile."""

    serie = pd.date_range(
        f"{year}-01-01", f"{year+1}-01-01", freq="h", inclusive="left"
    )

    das_scaling_factors = []

    ratios = profiles.ratios

    offset = 0
    for t in profiles.types:
        # create an array with the ratios
        t_len = t.size
        this_ratios = ratios[:, offset : offset + t_len]
        this_ratios = np.nan_to_num(this_ratios, nan=1.0 / t_len)
        offset += t_len
        # Create a data array for these ratios and convert to scaling factors
        da_sf = get_profile_da(profile=t(this_ratios), year=year) * t_len
        # Interpolate the data array
        da_interp = da_sf.interp(
            datetime=serie,
            method=interpolation_method,
            assume_sorted=True,
        )

        das_scaling_factors.append(da_interp.expand_dims("profile_type"))

    # Multiply the data arrays
    da = xr.concat(das_scaling_factors, dim="profile_type").prod(dim="profile_type")
    da_ratios = da / da.sum(dim="datetime")
    # Create the profile
    if return_profiles:
        return CompositeTemporalProfiles.from_ratios(
            da_ratios.values, types=[get_leap_year_or_normal(HourOfYearProfile, year)]
        )

    return da_ratios


def resolve_daytype(
    profiles: CompositeTemporalProfiles, profiles_indexes: xr.Dataset, year: int
) -> tuple[CompositeTemporalProfiles, xr.Dataset]:

    time_range = pd.date_range(
        f"{year}-01-01", f"{year+1}-01-01", freq="h", inclusive="left"
    )

    # Few checks on the day types given
    assert (
        "day_type" in profiles_indexes.dims
    ), "The profiles indexes must have a 'day_type' dimension"
    day_types = profiles_indexes.day_type.values
    specific_days = [SpecificDay(day_type) for day_type in day_types]
    days_mapping = {day: get_days_as_ints(day) for day in specific_days}
    all_values = sum([days for days in days_mapping.values()], [])
    if sorted(all_values) != sorted(list(range(7))):
        raise ValueError(
            f"Invalid {day_types=}, must cover all days of the week but they cover {all_values}."
        )

    # Check that the profile given is correct
    expected_profile = get_leap_year_or_normal(HourOfYearProfile, year)
    if not isinstance(profiles, CompositeTemporalProfiles | expected_profile):
        raise TypeError(
            f"{profiles=} must be a {CompositeTemporalProfiles} or {expected_profile}."
        )
    if profiles.types != [expected_profile]:
        raise ValueError(
            f"{profiles=} must contain only {expected_profile} for the given {year=}."
        )

    # Get only the profiles which are not the same on the day types
    dims = list(profiles_indexes.dims)
    dims.remove("day_type")
    stacked_indexes = profiles_indexes.stack(ind=dims)
    mask_differ_over_daytype = ~(
        stacked_indexes == stacked_indexes.isel(day_type=0)
    ).all("day_type")
    require_merge_indexes = stacked_indexes.loc[{"ind": mask_differ_over_daytype}]

    if np.any(require_merge_indexes == -1):
        raise ValueError(
            f"Cannot resolve {profiles_indexes=} as some profiles are missing. \n"
            "Please fill them with constant profiles."
        )

    # Get the ratios of each datetime and time
    da_ratios = xr.DataArray(
        profiles.ratios[require_merge_indexes],
        dims=["day_type", "ind", "time"],
        coords={
            "day_type": require_merge_indexes["day_type"],
            "ind": require_merge_indexes["ind"],
            "time": time_range,
        },
    )

    # Create the output index array
    da_indexes_out = xr.zeros_like(
        mask_differ_over_daytype.loc[mask_differ_over_daytype], dtype=int
    )

    da_profiles_out = da_ratios.isel(day_type=0).drop_vars("day_type")

    for day_type, days in days_mapping.items():
        # Assign the values of each datetime to the correct day
        mask = da_ratios.time.dt.dayofweek.isin(days)
        da_profiles_out.loc[{"time": mask}] = da_ratios.sel(
            day_type=day_type.value
        ).loc[{"time": mask}]

    new_profiles, new_indexes = ratios_dataarray_to_profiles(
        da_profiles_out.rename({"time": "ratio"}).unstack(fill_value=-1)
    )

    new_indexes = new_indexes.stack(ind=dims)
    # Remove the missing
    new_indexes = new_indexes.loc[new_indexes != -1]
    # Set them to the correct value  (as we later append at the end the new profiles)
    new_indexes += profiles.n_profiles

    out_indexes = stacked_indexes.isel(day_type=0).drop_vars("day_type").copy(deep=True)
    out_indexes.loc[{"ind": new_indexes.ind}] = new_indexes

    # Need to rescale as now the ratio might not sum up to 1 exactly
    new_profiles /= new_profiles.sum(axis=1).reshape(-1, 1)

    out_profiles = CompositeTemporalProfiles.from_ratios(
        np.concatenate([profiles.ratios, new_profiles]), types=[expected_profile]
    )

    return out_profiles, out_indexes.unstack("ind", fill_value=-1)


def ensure_specific_days_consistency(
    profiles: list[AnyTimeProfile],
) -> list[AnyTimeProfile]:
    """Make sure that there is not confilct between specific days profiles and normal daily profiles.

    In case there is any conflict, this return a profile for each day of the week.
    """

    if not any(isinstance(p, SpecificDayProfile) for p in profiles):
        return profiles

    # Get the specific days profiles

    daily_profiles = [p for p in profiles if isinstance(p, DailyProfile)]
    non_daily_profiles = [p for p in profiles if not isinstance(p, DailyProfile)]

    weekdays_profiles = {
        SpecificDay.MONDAY: None,
        SpecificDay.TUESDAY: None,
        SpecificDay.WEDNESDAY: None,
        SpecificDay.THURSDAY: None,
        SpecificDay.FRIDAY: None,
        SpecificDay.SATURDAY: None,
        SpecificDay.SUNDAY: None,
    }
    # First assign what we have sepcific
    for p in daily_profiles:
        if isinstance(p, SpecificDayProfile):
            days = days_of_specific_day(p.specific_day)
            if len(days) == 1:
                # This day was concerned specifically
                weekdays_profiles[p.specific_day] = p
            else:
                # This was a weekend or weekday profile
                for day in days:
                    # Only override if not already defined
                    if weekdays_profiles[day] is None:
                        weekdays_profiles[day] = SpecificDayProfile(
                            specific_day=day, ratios=p.ratios
                        )

    general_daily_profiles = [p for p in daily_profiles if type(p) == DailyProfile]

    # Add a constant profile for missing day
    for day, profile in weekdays_profiles.items():
        if profile is not None:
            continue
        if len(general_daily_profiles) == 0:
            p = SpecificDayProfile(specific_day=day)

        elif len(general_daily_profiles) == 1:
            p = SpecificDayProfile(
                specific_day=day, ratios=general_daily_profiles[0].ratios
            )
        else:
            raise ValueError(
                f"Cannot assign {general_daily_profiles=} to {day=}, more than one"
                f" general {type(DailyProfile)} was given."
            )
        weekdays_profiles[day] = p

    return non_daily_profiles + list(weekdays_profiles.values())


def create_scaling_factors_time_serie(
    start_time: datetime,
    end_time: datetime,
    profiles: list[AnyTimeProfile],
    apply_month_interpolation: bool = True,
    freq: str = "h",
    inclusive: str = "both",
    local_tz: str | None = None,
) -> pd.Series:
    """Create a time serie of ratios for the requested time range.

    :arg start_time: The start time of the time serie.
    :arg end_time: The end time of the time serie.
    :arg profiles: The profiles to use to create .
    :arg apply_month_interpolation: If True, apply the month interpolation.
    :arg inclusive: {“both”, “neither”, “left”, “right”}, default “both”
        same as `pd.date_range <https://pandas.pydata.org/docs/reference/api/pandas.date_range.html#pandas-date-range>`_
        Include boundaries; Whether to set each bound as closed or open.
    """

    kwargs = {}
    if local_tz is not None:
        kwargs["tz"] = "UTC"
    # Create the time serie
    time_serie = pd.date_range(
        start_time, end_time, freq=freq, inclusive=inclusive, **kwargs
    )
    if local_tz is not None:
        time_serie = time_serie.tz_convert(local_tz)

    # Create the scaling factors
    scaling_factors = np.ones(len(time_serie))

    # Correct profiles list with specific day profiles
    profiles = ensure_specific_days_consistency(profiles)

    # Apply the profiles
    for profile in profiles:
        scaling_factors *= profile_to_scaling_factors(
            time_serie, profile, apply_month_interpolation=apply_month_interpolation
        )

    return pd.Series(scaling_factors, index=time_serie)


def profile_to_scaling_factors(
    time_serie: pd.DatetimeIndex,
    profile: AnyTimeProfile,
    apply_month_interpolation: bool = True,
) -> np.ndarray:
    """Convert a temporal profile to a time serie.

    :arg apply_month_interpolation: If True, apply the month interpolation.
        Only applies when the profile is a :py:class: MounthsProfile
        Each mounthly values is assinged to the 15th of the month and then
        interpolation is used to get the values for the other days.

    :return: Scaling factors
        An array by which you can multiply the emission factor using the time serie.

    """

    # Get scaling factors, convert the ratios to scaling factors
    if len(profile) != 1:
        raise ValueError(
            f"Cannot apply {profile=} to a time serie, it must have only one profile."
        )

    factors = ratios_to_factors(profile.ratios.reshape(-1))

    # This will be the output
    scaling_factors = np.ones(len(time_serie))

    # Get the profile
    if isinstance(profile, SpecificDayProfile):
        # Find the days corresponding to this factor
        days_allowed = get_days_as_ints(profile.specific_day)
        if len(days_allowed) != 1:
            raise ValueError(
                f"Cannot apply {profile=} to a time serie, it must have only one day."
                "convert the time profiles with `ensure_specific_days_consistency`."
            )
        mask_matching_day = np.isin(time_serie.day_of_week, days_allowed)
        for hour, factor in enumerate(factors):
            # Other days will not have a scaling factor
            scaling_factors[(time_serie.hour == hour) & mask_matching_day] *= factor
    elif isinstance(profile, DailyProfile):
        # Get the mask for each hour of day and apply the scaling factor
        for hour, factor in enumerate(factors):
            scaling_factors[time_serie.hour == hour] *= factor
    elif isinstance(profile, WeeklyProfile):
        # Get the mask for each day of week and apply the scaling factor
        for day, factor in enumerate(factors):
            scaling_factors[time_serie.dayofweek == day] *= factor
    elif isinstance(profile, MounthsProfile):
        if apply_month_interpolation:
            # Apply the factor to the 15 th of each month by getting the exact datetime
            # of the 15th of each month
            mid_months = pd.date_range(
                start=time_serie[0],
                end=time_serie[-1],
                freq="MS",
            ) + timedelta(days=14)
            mid_months_factors = np.ones(len(mid_months))
            # Set the value to each month
            for month, factor in enumerate(factors):
                # Months start with 1
                month += 1
                mid_months_factors[mid_months.month == month] *= factor
            # Interpolate the values to the other days
            scaling_factors = np.interp(
                time_serie,
                mid_months,
                mid_months_factors,
            )
        else:
            # Simply apply to each month the scaling factor
            for month, factor in enumerate(factors):
                # Months start with 1
                month += 1
                scaling_factors[time_serie.month == month] *= factor
    else:
        raise NotImplementedError(
            f"Cannot apply {profile=}, {type(profile)=} is not implemented."
        )

    # Return the scaling factors
    return scaling_factors


_weekdays_long = [
    "Monday",
    "Tuesday",
    "Wednesday",
    "Thursday",
    "Friday",
    "Saturday",
    "Sunday",
]
_months_short = [
    "Jan",
    "Feb",
    "Mar",
    "Apr",
    "May",
    "Jun",
    "Jul",
    "Aug",
    "Sep",
    "Oct",
    "Nov",
    "Dec",
]
_months_long = [
    "January",
    "February",
    "March",
    "April",
    "May",
    "June",
    "July",
    "August",
    "September",
    "October",
    "November",
    "December",
]


timprofile_colnames = {
    WeeklyProfile: [
        _weekdays_short := [i[:3] for i in _weekdays_long],
        [i.lower() for i in _weekdays_short],
        _weekdays_long,
        [i.lower() for i in _weekdays_long],
        # TNO AVENGERS Format
        [i[:2] for i in _weekdays_short],
    ],
    MounthsProfile: [
        _months_short,
        [i.lower() for i in _months_short],
        _months_long,
        [i.lower() for i in _months_long],
    ],
    DailyProfile: [
        [str(i) for i in range(1, 25)],
        hours := [str(i) for i in range(24)],
        # TNO AVENGERS Format
        [f"H{i}" for i in hours],
    ],
}


def read_temporal_profiles(
    profiles_dir: PathLike,
    time_profiles_files_format: str = "timeprofiles*.csv",
    profile_csv_kwargs: dict[str, Any] = {},
    rtol: float = 1e-5,
) -> tuple[list[list[AnyTimeProfile]] | None, xr.DataArray | None]:
    """Read the temporal profiles csv files to the emiproc inventory format.

    The files for the time profiles are csv and must be all in the same directory
    named according to the argument `time_profiles_files_format`.

    Extra arguments depending on the file format can be passed to
    the function :py:func:`temporal_profiles.from_csv`
    that reads the csv files using: `profile_csv_kwargs`

    This returns the time profiles read, and and index xarray matching
    each substance and category to the desired profiles.

    If no files are found, this returns a warning.

    The format of the file will influence the name of the columns.
    Use the day of the weeks or the month names or the hour of the day to define the profile.

    :

    """

    # Note: The logic of this is a bit tricky because it has to handle
    #       the case where the profiles are speciated or not.

    profiles_dir = Path(profiles_dir)
    logger = logging.getLogger("emiproc.profiles.read_temporal_profiles")

    # List files with the expected format
    if profiles_dir.is_file():
        files = [profiles_dir]
        logger.info(f"File {profiles_dir=} found, will be used for timeprofiles.")
    else:
        if not profiles_dir.is_dir():
            raise ValueError(f"{profiles_dir=} is not a file or a directory.")
        files = list(profiles_dir.glob(time_profiles_files_format))
        if not files:
            logger.warning(
                "Cannot find any temporal profiles matching"
                f" {time_profiles_files_format=} in {profiles_dir=}.\n"
            )
            return None, None
        logger.info(
            f"Found {len(files)} files matching {time_profiles_files_format=} in"
            f" {profiles_dir=}"
        )

    out_profiles = AnyProfiles()
    indexes_list: list[xr.DataArray] = []
    for file in files:
        df = read_profile_file(file, **profile_csv_kwargs)
        possible_matching = {
            profile_type: colnames
            for profile_type, colnames_list in timprofile_colnames.items()
            for colnames in colnames_list
            if all(col in df.columns for col in colnames)
        }
        if not possible_matching:
            raise ValueError(
                f"Cannot find any matching time profile for {file=} with Columns"
                f" {df.columns}."
                "Please check the file format."
                "See more about time profiles file at "
                "https://emiproc.rtfd.io/en/latest/api.html#emiproc.profiles.temporal_profiles.read_temporal_profiles"
            )
        logger.info(f"{possible_matching=}")
        # Generate the profiles objects
        indexes = get_profiles_indexes(df)
        for profile_type, colnames in possible_matching.items():
            try:
                ratios = np.array([df[col] for col in colnames])
                if np.all(np.isclose(ratios.sum(axis=0), 1.0, rtol=rtol)):
                    # Ratios found
                    ratios = ratios
                elif np.all(np.isclose(np.mean(ratios, axis=0), 1.0, rtol=rtol)):
                    # Scaling factors found
                    ratios = ratios / ratios.sum(axis=0)
                else:
                    raise ValueError(
                        "Could not determine if scaling factors or ratios were given"
                        f" in {file=}.\n data:{ratios=} and \n"
                        f" mean:{np.mean(ratios, axis=0)} \n"
                        f" sum:{np.sum(ratios, axis=0)} \n Try to set {rtol=} to a"
                        " higher value if this is due to rounding erros."
                    )
                if isinstance(profile_type, tuple):
                    profiles = profile_type[0](
                        ratios=ratios.T, specific_day=profile_type[1]
                    )
                else:
                    profiles = profile_type(ratios.T)
            except Exception as e:
                raise ValueError(
                    f"Cannot create profile {profile_type=} from {file=} with {ratios=}"
                ) from e
            indexes += len(out_profiles)
            out_profiles.append(profiles)
            # Add a new dim which is the profile type
            indexes_list.append(
                indexes.expand_dims({"profile": [profile_type.__name__]})
            )

    combined_indexes = merge_indexes(indexes_list)

    composite_profiles, out_indexes = make_composite_profiles(
        out_profiles, combined_indexes
    )
    # Drop the profile dim
    if "profile" in out_indexes.dims:
        out_indexes = out_indexes.drop_vars("profile")

    return composite_profiles, out_indexes


@emiproc.deprecated
def from_csv(
    file: PathLike,
    profile_csv_kwargs: dict[str, Any] = {},
) -> dict[str | tuple[str, str] : AnyTimeProfile]:
    """Create the profile from a csv file.

    Based on the file, guess the correct profile
    to create.

    This is now deprectaed. Use :py:func:`read_temporal_profiles` instead.
    """

    df, cat_header, sub_header = read_profile_csv(file, **profile_csv_kwargs)

    if "Mon" in df.columns:
        # Weekly profile with 3 letters identification
        data_columns = ["Mon", "Tue", "Wed", "Thu", "Fri", "Sat", "Sun"]
        profile = WeeklyProfile
    elif "mon" in df.columns:
        # Weekly profile with 3 letters identification
        data_columns = ["mon", "tue", "wed", "thu", "fri", "sat", "sun"]
        profile = WeeklyProfile
    elif " mon " in df.columns:
        data_columns = [" mon ", " tue ", " wed ", " thu ", " fri ", " sat ", " sun "]
        profile = WeeklyProfile
    elif "Monday" in df.columns:
        # Weekly profile with full name identification
        data_columns = [
            "Monday",
            "Tuesday",
            "Wednesday",
            "Thursday",
            "Friday",
            "Saturday",
            "Sunday",
        ]
        profile = WeeklyProfile
    elif "Jan" in df.columns:
        # Yearly profile with 3 letters identification
        data_columns = [
            "Jan",
            "Feb",
            "Mar",
            "Apr",
            "May",
            "Jun",
            "Jul",
            "Aug",
            "Sep",
            "Oct",
            "Nov",
            "Dec",
        ]
        profile = MounthsProfile
    elif " jan " in df.columns:
        data_columns = [
            " jan ",
            " feb ",
            " mar ",
            " apr ",
            " may ",
            " jun ",
            " jul ",
            " aug ",
            " sep ",
            " oct ",
            " nov ",
            " dec ",
        ]
        profile = MounthsProfile
    elif "January" in df.columns:
        # Yearly profile with full name identification
        data_columns = [
            "January",
            "February",
            "March",
            "April",
            "May",
            "June",
            "July",
            "August",
            "September",
            "October",
            "November",
            "December",
        ]
        profile = MounthsProfile
    elif (
        "0" not in df.columns
        and "1" in df.columns
        and "24" in df.columns
        and "25" not in df.columns
    ):
        # Daily profile
        data_columns = [str(i) for i in range(1, 25)]
        profile = DailyProfile
    else:
        raise NotImplementedError(f"Cannot guess the profile from {file=}")

    # Check that all the data columns are in the file
    for col in data_columns:
        if not col in df.columns:
            raise ValueError(
                f"Cannot find {col=} in {file=}, required to build {profile=}"
            )

    # Read all rows to generate the profiles
    profiles = {}
    for _, row in df.iterrows():
        cat = row[cat_header]
        if sub_header is not None:
            sub = row[sub_header]
            key = (cat, sub)
        else:
            key = cat
        ratios = np.array([row[col] for col in data_columns])

        # Check if scaling factors are given instead of ratios
        if np.isclose(ratios.sum(), len(ratios)):
            ratios = ratios / ratios.sum()
        elif not np.isclose(ratios.sum(), 1.0):
            pass  # Already ratios
        else:
            raise ValueError(
                f"Cannot guess if {ratios=} are ratios or scaling factors for {row=} in"
                f" {file=}."
            )
        profiles[key] = profile(ratios=ratios)

    return profiles


def from_yaml(yaml_file: PathLike) -> list[AnyTimeProfile]:
    """Read a yml file containing a temporal profile.

    Only one temporal profile is currently accepted in the yaml definition.
    """
    logger = logging.getLogger("emiproc.profiles.from_yaml")
    yaml_file = Path(yaml_file)

    with open(yaml_file, "r") as f:
        data = yaml.safe_load(f)

    if data is None:
        logger.warning(f"Empty yaml file {yaml_file=}")
        return []
    elif not isinstance(data, dict):
        raise ValueError(f"Invalid yaml file {yaml_file=}, expected to load a dict.")

    profiles = []

    # Create possible aliases for the names
    profiles_mapping: dict[AnyTimeProfile, list[str]] = {
        DailyProfile: ["diurn", "daily", "day"],
        SpecificDayProfile: [
            "diurn_weekday",
            "diurn_weekend",
            "diurn_saturday",
            "diurn_sunday",
        ],
        WeeklyProfile: ["weekly", "week"],
        MounthsProfile: ["season", "year", "monthly", "month"],
    }
    profile_of_key = {
        key: profile for profile, keys in profiles_mapping.items() for key in keys
    }

    _types_added = []
    # Check that the yaml does not contain any unkown key
    for key in data.keys():
        if not any(key in profile_names for profile_names in profiles_mapping.values()):
            logger.warning(f"Unknown key {key=} in {yaml_file=}")
            continue
        profile_class = profile_of_key[key]
        # Check that a profile of that type was not already added
        if profile_class in _types_added:
            raise ValueError(
                f"Cannot add {key=} to {yaml_file=} as a {profile_class=} was already"
                " added."
            )
        # add the profile
        ratios = data[key]
        # Check the ratio
        if not np.isclose(np.sum(ratios), 1.0):
            raise ValueError(
                f"{ratios=} in {yaml_file=} do not sum to 1 but {np.sum(ratios)}."
            )

        # Add additional information on the profiles if requried
        kwargs = {}
        if profile_class is SpecificDayProfile:
            # get the type of the profile
            profile_type = key.split("_")[-1]
            # Add the selected day
            kwargs["specific_day"] = SpecificDay(profile_type)

        try:
            profile = profile_class(ratios=ratios, **kwargs)
        except Exception as e:
            raise ValueError(
                f"Cannot create profile {key=} from {yaml_file=} with {ratios=}"
            ) from e

        profiles.append(profile)

    if len(profiles) == 0:
        logger.warning(f"No profile found in {yaml_file=}")
    return profiles


def to_yaml(profiles: list[AnyTimeProfile], yaml_file: PathLike):
    """Write a list of profiles to a yaml file."""
    yaml_file = Path(yaml_file)
    yaml_file.parent.mkdir(parents=True, exist_ok=True)

    data = {}
    for profile in profiles:
        if isinstance(profile, DailyProfile):
            if isinstance(profile, SpecificDayProfile):
                key = f"diurn_{profile.specific_day.value}"
            else:
                key = "daily"
        elif isinstance(profile, WeeklyProfile):
            key = "weekly"
        elif isinstance(profile, MounthsProfile):
            key = "monthly"
        else:
            raise NotImplementedError(f"Cannot write {profile=}")

        data[key] = profile.ratios.tolist()

    with open(yaml_file, "w") as f:
        yaml.dump(data, f)



"""Provides default values for how the data columns can be named."""

from datetime import datetime


accepted_category_colnames = [
    "Category",
    "GNFR_Category",
    "TNO GNFR sectors Sept 2018",
    "GNFR",
]
substances_colnames = [
    "Substance",
    "POLL",  # TNO
]
accepted_country_colnames = [
    "Country",
    "ISO3",  # TNO
]

metadata_colnames = [
    "GNFR_Category_Name",
]
type_colnames = [
    "Type",
]

time_colnames = [
    "Year",
    "Month",
    "Day",
    "Hour",
]

cell_colnames = ["Cell"]

daytype_colnames = ["DayType"]

attributes_accepted_colnames = {
    "category": accepted_category_colnames,
    "substance": substances_colnames,
    "cell": cell_colnames,
    "time": time_colnames,
    "country": accepted_country_colnames,
    "type": type_colnames,
    "day_type": daytype_colnames,
}

# Add lowercase versions of the accepted colnames
for key, value in attributes_accepted_colnames.items():
    attributes_accepted_colnames[key].extend([v.lower() for v in value])

type_of_dim = {
    "category": str,
    "substance": str,
    "cell": int,
    "time": datetime,
    "country": str,
    "type": str,
    "day_type": str,
}

all_reserved_colnames = sum(
    [
        accepted_category_colnames,
        substances_colnames,
        accepted_country_colnames,
        metadata_colnames,
        type_colnames,
        daytype_colnames,
    ],
    [],
)



"""Script used for creating raster data of zurich for the ICOS-cities project.

The idea of this scripts is to produce raster files for zurich.

It is possible put the rasters inside the swiss inventory as well.
"""

# %%
# autoreload modules in interactive python
%load_ext autoreload
%autoreload 2
# %%
from datetime import datetime
from enum import Enum
from math import floor
from pathlib import Path

import geopandas as gpd
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import xarray as xr
from shapely.geometry import Point, Polygon

from emiproc.exports.netcdf import nc_cf_attributes
from emiproc.grids import LV95, WGS84, SwissGrid
from emiproc.inventories.swiss import SwissRasters
from emiproc.inventories.utils import (
    add_inventories,
    crop_with_shape,
    get_total_emissions,
    scale_inventory,
    group_categories
)
from emiproc.inventories.zurich import MapLuftZurich
from emiproc.regrid import remap_inventory
from emiproc.speciation import speciate, speciate_inventory
from emiproc.utilities import SEC_PER_YR

# %% define some parameters for the output


YEAR = 2020

INCLUDE_SWISS_OUTSIDE = True
swiss_data_path = Path(
    r"C:\Users\coli\Documents\ZH-CH-emission\Data\CHEmissionen\CH_Emissions_2015_2020_2022_CO2_CO2biog_CH4_N2O_BC_AP.xlsx"
)
outdir = Path(r"C:\Users\coli\Documents\ZH-CH-emission\output_files\mapluft_rasters")
mapluft_dir = Path(r"C:\Users\coli\Documents\Data\mapluft_emissionnen_kanton")
mapluf_file = mapluft_dir / f"mapLuft_{YEAR}_v2024.gdb"

# edge of the raster cells
RASTER_EDGE = 100
VERSION = "v1.8"

# Whether to split the biogenic CO2 and the antoropogenic CO2
SPLIT_BIOGENIC_CO2 = True

# Whether to add the human respiration
ADD_HUMAN_RESPIRATION = True
# File with the data required for the human respiration
quartier_anlyse_dir = Path(r"C:\Users\coli\Documents\Data\Quartieranalyse_zurich") / str(YEAR)
quartier_anlyse_file = quartier_anlyse_dir / "Quartieranalyse_-OGD.gpkg"


# Make a Enum class for output unit choices
class Unit(Enum):
    """Enum class for units"""

    kg_yr = "kg yr-1 cell-1"
    ug_m2_s = "μg m-2 s-1"


output_unit = Unit.kg_yr

# Whether to group categories to the GNRF categories
USE_GNRF = True


# Whether to split the F category of the GNRF into 4 subcategories for accounting
# for the different vehicle types (cars, light duty, heavy duty, two wheels)
SPLIT_GNRF_ROAD_TRANSPORT = True


# %% Check some parameters and create the output directory
weights_dir = outdir / f"weights_files_{RASTER_EDGE}_{YEAR}_{VERSION}"

if SPLIT_GNRF_ROAD_TRANSPORT and not USE_GNRF:
    raise ValueError("Cannot split GNRF if not using GNRF")

if INCLUDE_SWISS_OUTSIDE:
    # Swiss inventory code works only if the raster is the same as the swiss raster (100 m )
    assert RASTER_EDGE == 100
    # Need to have the same categories between swiss and zurich
    assert USE_GNRF


# %% load the zurich inventory
inv = MapLuftZurich(mapluf_file)


# %%
def load_zurich_shape(
    zh_raw_file=r"C:\Users\coli\Documents\ZH-CH-emission\Data\Zurich_borders.txt",
    crs_file: int = WGS84,
    crs_out: int = LV95,
) -> Polygon:
    with open(zh_raw_file, "r") as f:
        points_list = eval(f.read())
        zh_poly = Polygon(points_list[0])
        zh_poly_df = gpd.GeoDataFrame(geometry=[zh_poly], crs=crs_file).to_crs(crs_out)
        zh_poly = zh_poly_df.geometry.iloc[0]
        return zh_poly


zh_shape = load_zurich_shape()

x_min, y_min, x_max, y_max = zh_shape.bounds

# %% create the zurich swiss grid

d = RASTER_EDGE
xmin, ymin = floor(x_min) // d * d, floor(y_min) // d * d
nx, ny = int(x_max - xmin) // d, int(y_max - ymin) // d
xs = np.arange(xmin, xmin + nx * d, step=d)
ys = np.arange(ymin, ymin + ny * d, step=d)
point_x = np.repeat(xs, ny)
point_y = np.tile(ys, nx)

polys = [
    Polygon(((x, y), (x + d, y), (x + d, y + d), (x, y + d)))
    for y in reversed(ys)
    for x in xs
]
centers = [Point((x + d / 2, y + d / 2)) for y in reversed(ys) for x in xs]
zh_gdf = gpd.GeoDataFrame(geometry=polys, crs=LV95)
gdf_centers = gpd.GeoDataFrame(geometry=centers, crs=LV95)
# %% prepare the output on WGS84
WGS84_gdf = zh_gdf.to_crs(WGS84)
for i in range(4):
    WGS84_gdf[f"coord_{i}"] = WGS84_gdf.geometry.apply(
        lambda poly: poly.exterior.coords[i]
    )

# %% Split the biogenic CO2

if SPLIT_BIOGENIC_CO2:
    from emiproc.inventories.zurich.speciation_co2_bio import ZH_CO2_BIO_RATIOS

    inv = speciate(inv, "CO2", ZH_CO2_BIO_RATIOS, drop=True)

# %% do the actual remapping of zurich to rasters

rasters_inv = remap_inventory(
    crop_with_shape(inv, zh_shape),
    zh_gdf.geometry,
    weigths_file=weights_dir
    / f"{mapluf_file.stem}_weights",
)


# %% change the categories
if USE_GNRF:
    
    from emiproc.inventories.zurich.gnrf_groups import ZH_2_GNFR

    if SPLIT_GNRF_ROAD_TRANSPORT:
        ZH_2_GNFR = ZH_2_GNFR.copy()
        # Remove the road transport from the GNRF
        ZH_2_GNFR.pop("GNFR_F")
        splitted_cats = {
            "GNFR_F-cars": [
                "c1301_Personenwagen_Emissionen_Kanton",
                "c1306_StartStopTankatmung_Emissionen_Kanton",
            ],
            "GNFR_F-light_duty": [
                "c1307_Lieferwagen_Emissionen_Kanton",
            ],
            "GNFR_F-heavy_duty": [
                "c1302_Lastwagen_Emissionen_Kanton",
                "c1304_Linienbusse_Emissionen_Kanton",
                "c1305_Trolleybusse_Emissionen_Kanton",
                "c1308_Reisebusse_Emissionen_Kanton",
            ],
            "GNFR_F-two_wheels": [
                "c1303_Motorraeder_Emissionen_Kanton",
            ],
        }
        # add this to the mapping
        ZH_2_GNFR |= splitted_cats

    rasters_inv = group_categories(rasters_inv, ZH_2_GNFR)

# %% add the swiss inventory when needed
if INCLUDE_SWISS_OUTSIDE:
    inv_ch = SwissRasters(
        data_path=swiss_data_path,
        rasters_dir=swiss_data_path.parent / "ekat_gridascii",
        rasters_str_dir=swiss_data_path.parent / "ekat_str_gridascii_v17",
        requires_grid=True,
        # requires_grid=False,
        year=YEAR,
        # Specify the compound in the inventory and how they should be named in the output
        dict_spec={
            "NOX": "NOx",
            "PM2.5": "PM25",
            "F-Gase": "F-gases",
            "CO2": "CO2_fos",
            "CO2_biog": "CO2_bio",
        },
    )

    inv_ch.history.append(
        "the map of CO2 for evstr was used for BC and CO2-bio as they did not exist"
    )

    from emiproc.inventories.categories_groups import CH_2_GNFR
    
    # These categories are not in the invenotry here, because we don't care about them
    missing_cats = ['eilgk', 'evklm', 'evtrk', 'enwal']
    # Remove the missing categories
    our_CH_2_GNFR = {
        new_cat: [c for c in cats if c not in missing_cats] for new_cat, cats in CH_2_GNFR.items() 
    }
    groupped_ch = group_categories(inv_ch, our_CH_2_GNFR)



    if SPLIT_GNRF_ROAD_TRANSPORT:
        # Calculate splitting ratios in zurich
        total_emisson = get_total_emissions(rasters_inv)
        speciation_dict = {}
        for sub, cat_dic in total_emisson.items():
            # Get the categories of the GNRF-F
            f_cat_dict = {
                cat: val for cat, val in cat_dic.items() if cat.startswith("GNFR_F")
            }
            # Get the total of the GNRF-F
            f_total = sum(f_cat_dict.values())
            # Calculate the ratios
            catsub = ("GNFR_F", sub)
            if catsub in groupped_ch._gdf_columns:
                speciation_dict[catsub] = {
                    (cat, sub): val / f_total for cat, val in f_cat_dict.items()
                }

        groupped_ch = speciate_inventory(groupped_ch, speciation_dict, drop=True)

    ch_outside_zh = crop_with_shape(
        groupped_ch,
        zh_shape,
        keep_outside=True,
        modify_grid=False,
        weight_file=weights_dir / "ch_out_zh",
    )
    ch_inside_zh = crop_with_shape(
        groupped_ch,
        zh_shape,
        keep_outside=False,
        modify_grid=False,
        weight_file=weights_dir / "ch_in_zh",
    )


# %%
if INCLUDE_SWISS_OUTSIDE:
    remapped_ch_out = remap_inventory(
        ch_outside_zh,
        zh_gdf.geometry,
        # weigths_file=(weights_dir / f"swiss_around_zh_2_{RASTER_EDGE}x{RASTER_EDGE}"),
    )
# %% Rescale the swiss and add it, the scaling is made such that the
# mapluft inventory is not changed and the total swiss inventory is also not changed
# so we only scale the region outside of zurich to compensate
if INCLUDE_SWISS_OUTSIDE:
    # get the total inside zurich from mapluft
    mapluft_total = get_total_emissions(rasters_inv)
    # get the total inside zurich from swiss inv
    swiss_out_total = get_total_emissions(ch_outside_zh)
    swiss_total = get_total_emissions(groupped_ch)
    # calculates scalings
    scaling_factors = {}
    for sub, cat_dic in swiss_total.items():
        if sub not in mapluft_total.keys():
            continue
        scaling_factors[sub] = {}
        for cat, total in cat_dic.items():
            if cat == "__total__" or cat not in mapluft_total[sub]:
                continue
            # we want scaling_factor * swiss_out + mapluft = swiss_total
            scaling_factor = (total - mapluft_total[sub][cat]) / swiss_out_total[sub][
                cat
            ]
            scaling_factors[sub][cat] = scaling_factor
    # rescale inventory
    rescaled_ch = scale_inventory(remapped_ch_out, scaling_factors)
    # add the inventories
    rasters_inv = add_inventories(rasters_inv, rescaled_ch)

# %% Add the human respiration
if ADD_HUMAN_RESPIRATION:

    from emiproc.human_respiration import (
        load_data_from_quartieranalyse,
        people_to_emissions,
        EmissionFactor,
    )

    # Load the data. It is available for the whole Kanton of zurich, 
    # which covers the whole grid of the output
    df_quariter = load_data_from_quartieranalyse(quartier_anlyse_file)
    # Load into an emiproc Inventory
    raw_resp_inv = people_to_emissions(
        df_quariter,
        # Assumes people spend 60% of their time at home and 40% at work
        time_ratios={"people_living": 0.6, "people_working": 0.4},
        emission_factor={
            ("people_living", "CO2_bio"): EmissionFactor.ROUGH_ESTIMATON,
            ("people_working", "CO2_bio"): EmissionFactor.ROUGH_ESTIMATON,
            ("people_living", "N2O"): EmissionFactor.N2O_MITSUI_ET_ALL,
            ("people_working", "N2O"): EmissionFactor.N2O_MITSUI_ET_ALL,
            ("people_living", "CH4"): EmissionFactor.CH4_POLAG_KEPPLER,
            ("people_working", "CH4"): EmissionFactor.CH4_POLAG_KEPPLER,
        },
    )

    # Group the categories
    resp_inv = group_categories(
        raw_resp_inv,
        {
            "GNFR_O": ["people_living", "people_working"],
        },
    )
    # If keep inside, crop the inventory to the zurich shape
    if not INCLUDE_SWISS_OUTSIDE:
        resp_inv = crop_with_shape(resp_inv, zh_shape)

    # Remap the inventory to the raster
    remapped_resp = remap_inventory(
        resp_inv,
        zh_gdf.geometry,
        weigths_file=weights_dir / f"resp_weights_{INCLUDE_SWISS_OUTSIDE}",
    )

    rasters_inv = add_inventories(rasters_inv, remapped_resp)
# %% Populate the dataframe of the output


ds_out = xr.Dataset(
    coords={
        "x": (
            "x",
            xs + d / 2,
            {
                "standard_name": "easting",
                "long_name": "easting",
                "units": "m",
                "comment": "center_of_cell",
                "projection": "Swiss coordinate system LV95",
            },
        ),
        "y": (
            "y",
            (ys + d / 2)[::-1],
            {
                "standard_name": "northing",
                "long_name": "northing",
                "units": "m",
                "comment": "center_of_cell",
                "projection": "Swiss coordinate system LV95",
            },
        ),
        "lon": (
            ("y", "x"),
            gdf_centers.to_crs(WGS84)
            .geometry.apply(lambda point: point.coords[0][0])
            .to_numpy()
            .reshape((ny, nx)),
            {
                "standard_name": "longitude",
                "long_name": "longitude",
                "units": "degrees_east",
                "comment": "center_of_cell",
                "bounds": "lon_bnds",
                "projection": "WGS84",
            },
        ),
        "lat": (
            ("y", "x"),
            gdf_centers.to_crs(WGS84)
            .geometry.apply(lambda point: point.coords[0][1])
            .to_numpy()
            .reshape((ny, nx)),
            {
                "standard_name": "latitude",
                "long_name": "latitude",
                "units": "degrees_north",
                "comment": "center_of_cell",
                "bounds": "lat_bnds",
                "projection": "WGS84",
            },
        ),
        "lon_bnds": (
            ("nv", "y", "x"),
            np.array(
                [
                    WGS84_gdf[f"coord_{i}"]
                    .apply(lambda p: p[0])
                    .to_numpy()
                    .reshape((ny, nx))
                    for i in range(4)
                ]
            ),
            {
                "comment": "cell boundaries, anticlockwise",
            },
        ),
        "lat_bnds": (
            ("nv", "y", "x"),
            np.array(
                [
                    WGS84_gdf[f"coord_{i}"]
                    .apply(lambda p: p[1])
                    .to_numpy()
                    .reshape((ny, nx))
                    for i in range(4)
                ]
            ),
            {
                "comment": "cell boundaries, anticlockwise",
            },
        ),
        "source_category": ("source_category", rasters_inv.categories, {}),
    },
    data_vars={
        f"emi_{sub}": (
            ("source_category", "y", "x"),
            np.full((len(rasters_inv.categories), ny, nx), np.nan),
            {
                "standard_name": (
                    f"tendency_of_atmosphere_mass_content_of_{sub}_due_to_emission"
                ),
                "long_name": f"Emissions of {sub}",
                "units": output_unit.value,
                "comment": "annual mean emission rate",
            },
        )
        for sub in rasters_inv.substances
    }
    | {
        f"emi_{sub}_total": (
            "source_category",
            np.full(len(rasters_inv.categories), np.nan),
            {
                "long_name": f"Total Emissions of {sub}",
                "units": "kg yr-1",
            },
        )
        for sub in rasters_inv.substances
    }
    | {
        f"emi_{sub}_all_sectors": (
            ("y", "x"),
            np.zeros((ny, nx)),
            {
                "standard_name": (
                    f"tendency_of_atmosphere_mass_content_of_{sub}_due_to_emission"
                ),
                "long_name": f"Aggregated Emissions of {sub} from all sectors",
                "units": output_unit.value,
                "comment": "annual mean emission rate",
            },
        )
        for sub in rasters_inv.substances
    },
    attrs=nc_cf_attributes(
        author="Lionel Constantin, Empa",
        contact="dominik.brunner@empa.ch",
        title=(
            "Annual mean emissions of CO2 of the city of Zurich (only emissions within"
            " the political borders of the city)"
        ),
        source="https://www.stadt-zuerich.ch/gud/de/index/umwelt_energie/luftqualitaet/schadstoffquellen/emissionskataster.html",
        comment="Created for use in the EU project ICOS-Cities",
        history=(
            "Created from original GIS inventory mapLuft of the city of Zurich by"
            " rasterizing all point, line and area sources"
        ),
        additional_attributes={
            "swiss_coordinate_system_lv95": "https://www.swisstopo.admin.ch/en/knowledge-facts/surveying-geodesy/coordinates/swiss-coordinates.html",
            "comment_lv95": (
                "In original LV95 system, x denote northings and y eastings. They have"
                " been exchanged here for better compatibility with lon/lat."
            ),
            "copyright_notice": "",
            "script_version": VERSION,
            "emiproc_history": str(rasters_inv.history),
        },
    ),
)


ds_out
# %%
for category, sub in rasters_inv._gdf_columns:
    if (category, sub) not in rasters_inv._gdf_columns:
        continue
    emissions = rasters_inv.gdf[(category, sub)].to_numpy().reshape((ny, nx))
    # convert from kg/y to μg m-2 s-1
    # Get the desired unit
    if output_unit == Unit.ug_m2_s:
        rescaled = emissions * 1e9 / SEC_PER_YR / (RASTER_EDGE * RASTER_EDGE)
    elif output_unit == Unit.kg_yr:
        rescaled = emissions
    else:
        raise ValueError(f"Unknown unit {output_unit}")

    # Assign emissions of this category
    ds_out[f"emi_{sub}"].loc[dict(source_category=category)] = rescaled
    # Add the the categories aggregated emission
    ds_out[f"emi_{sub}_all_sectors"] += rescaled
    # Add to the total emission value
    ds_out[f"emi_{sub}_total"].loc[dict(source_category=category)] = np.sum(emissions)

# %%
ds_out.to_netcdf(
    outdir
    / f"zurich_{'inside_swiss' if INCLUDE_SWISS_OUTSIDE else 'cropped'}_{'Fsplit' if SPLIT_GNRF_ROAD_TRANSPORT else ''}_{RASTER_EDGE}x{RASTER_EDGE}_{mapluf_file.stem}_{VERSION}.nc"
)

# %%

print(f"Output written to {outdir}")

# %%



import numpy as np
import geopandas as gpd
import xarray as xr

from shapely.geometry import Polygon, Point

from emiproc.inventories import Inventory

from emiproc.profiles.vertical_profiles import (
    VerticalProfile,
    VerticalProfiles,
)


def get_random_profiles(
    num: int, heights: list[int] = [14, 50, 89]
) -> VerticalProfiles:
    """Get random profiles for testing."""
    ratios = np.random.rand(num, len(heights))
    return VerticalProfiles(ratios / ratios.sum(axis=1)[:, None], height=heights)


# Create test profiles
list_of_three = [
    VerticalProfile(np.array([0, 0.3, 0.7, 0.0]), np.array([15, 30, 60, 100])),
    VerticalProfile(
        np.array([0.1, 0.3, 0.5, 0.0, 0.1]), np.array([10, 30, 40, 65, 150])
    ),
    VerticalProfile(np.array([1]), np.array([20])),
]
VerticalProfiles_instance = VerticalProfiles(
    np.array(
        [
            [0.0, 0.3, 0.7, 0.0],
            [0.1, 0.2, 0.7, 0.0],
            [0.0, 0.3, 0.2, 0.5],
        ]
    ),
    np.array([15, 30, 60, 100]),
)

# Create a test geodataframe
gdf = gpd.GeoDataFrame(
    {
        ("test_cat", "CO2"): [i for i in range(4)],
        ("test_cat", "CH4"): [i + 3 for i in range(4)],
        # ("test_cat", "NH3"): [2 * i for i in range(4)],
        ("test_cat2", "CO2"): [i + 1 for i in range(4)],
        ("test_cat2", "CH4"): [i + 1 for i in range(4)],
        # ("test_cat2", "NH3"): [i + 1 for i in range(4)],
    },
    geometry=gpd.GeoSeries(
        [
            Polygon(((0, 0), (0, 1), (1, 1), (1, 0))),
            Polygon(((0, 1), (0, 2), (1, 2), (1, 1))),
            Polygon(((1, 0), (1, 1), (2, 1), (2, 0))),
            Polygon(((1, 1), (1, 2), (2, 2), (2, 1))),
        ]
    ),
)

# Corresponding profiles integer array
# -1 is when the profile is not defined
corresponding_vertical_profiles = xr.DataArray(
    [
        [[0, -1, 0, 2], [0, 2, 0, 0], [0, 0, 0, 0]],
        [[0, -1, -1, 1], [0, 0, -1, 0], [0, 1, 0, 0]],
    ],
    dims=("category", "substance", "cell"),
    coords={
        "category": ["test_cat", "test_cat2"],
        "substance": (substances := ["CH4", "CO2", "NH3"]),
        "cell": (cells := [0, 1, 2, 3]),
    },
)
corresponding_2d_profiles = xr.DataArray(
    [
        [2, 1, 0],
        [-1, 1, 0],
    ],
    dims=("category", "substance"),
    coords={
        "category": ["test_cat", "test_cat2"],
        "substance": (substances := ["CH4", "CO2", "NH3"]),
    },
)
single_dim_profile_indexes = xr.DataArray(
    [0, 1, -1],
    dims="category",
    coords={
        "category": ["test_cat", "test_cat2", "test_cat3"],
    },
)
single_dim_weights = xr.DataArray(
    [0.5, 0.2, 0.0],
    dims="category",
    coords={
        "category": ["test_cat", "test_cat2", "test_cat3"],
    },
)


inv = Inventory.from_gdf(
    gdf,
    # Add some point sources
    gdfs={
        "test_cat": gpd.GeoDataFrame(
            {
                "CO2": [1, 2, 3],
                "__v_profile__": [-1, 2, 1],
            },
            geometry=[Point(0.75, 0.75), Point(0.25, 0.25), Point(1.2, 1)],
        ),
        "test_cat2": gpd.GeoDataFrame(
            {
                "CO2": [1.2, 2.7, 8],
                "CH4": [4, 2, 8],
                "__v_profile__": [1, 2, -1],
            },
            geometry=[Point(0.65, 0.75), Point(1.1, 0.8), Point(1.2, 1)],
        ),
        "test_cat3": gpd.GeoDataFrame(
            {"CO2": [1, 2]},
            geometry=[Point(0.65, 0.75), Point(1.1, 0.8)],
        ),
    },
)
inv_groups_dict = {"new_cat": ["test_cat"], "new_cat2": ["test_cat2", "test_cat3"]}
inv_groups_subs_dict = {"co2": ["CO2"], "others": ["CH4", "NH3"]}

inv.v_profiles = VerticalProfiles_instance
inv.v_profiles_indexes = corresponding_vertical_profiles



from __future__ import annotations
from os import PathLike
from pathlib import Path
import xarray as xr
import numpy as np
from emiproc.inventories import Inventory
from emiproc.grids import RegularGrid
from emiproc.regrid import remap_inventory
from emiproc.exports.netcdf import NetcdfAttributes, nc_cf_attributes
from emiproc.utilities import Units, SEC_PER_YR, PER_CELL_UNITS, PER_M2_UNITS


def export_raster_netcdf(
    inv: Inventory,
    path: PathLike,
    grid: RegularGrid | None = None,
    netcdf_attributes: NetcdfAttributes | None = None,
    weights_path: PathLike | None = None,
    lon_name: str = "lon",
    lat_name: str = "lat",
    var_name_format: str | None = None,
    unit: Units = Units.KG_PER_YEAR,
    group_categories: bool = False,
    add_totals: bool = True,
) -> Path:
    """Export the inventory to a netcdf file as a raster.

    This will first remap the invenotry to a raster file using
    :py:func:`emiproc.regrid.remap_inventory` and
    then export the result to a netcdf file.

    :param inv: the inventory to export
    :param path: the path to the output file
    :param grid: the raster grid to export to
        If not given, assumes the inventory is already on a raster grid.
    :param netcdf_attributes: NetCDF attributes to add to the file.
        These can be generated using
        :py:func:`emiproc.exports.netcdf.nc_cf_attributes` .
    :param weights_path: Optionally,
        The path to the weights file to use for regridding.
        If not given, the weights will be calculated on the fly.
    :param lon_name: The name of the longitude dimension in the nc file.
    :param lat_name: The name of the latitude dimension in the nc file.
    :param var_name_format: The format string to use for the variable names.
        The format string should contain two named fields: ``substance`` and ``category``.
    :param unit: The unit of the emissions.
    :param group_categories: If True the categories will be grouped in the output file.
        Intead of a variable for each pair (substance, category) there will be a variable
        only for each substance and the categories will be grouped in a dimension.
    :param add_totals: If True, the total emissions for each substance will be added
        as new variables to the file.
        One will be the raster sum of all categories and the other will be the total sum
        over all cells.

    """

    if var_name_format is None:
        if group_categories:
            var_name_format = "{substance}"
        else:
            var_name_format = "{substance}_{category}"

    if grid is None:
        grid = inv.grid
    if grid != inv.grid:
        inv = remap_inventory(inv, grid, weights_path)

    if netcdf_attributes is None:
        netcdf_attributes = nc_cf_attributes()
    # add the history
    netcdf_attributes["emiproc_history"] = str(inv.history)

    crs = grid.crs

    if unit == Units.KG_PER_YEAR:
        conversion_factor = 1.0
    elif unit == Units.KG_PER_M2_PER_S:
        conversion_factor = (
            1 / SEC_PER_YR / np.array(grid.cell_areas).reshape(grid.shape).T
        )
    elif unit == Units.MUG_PER_M2_PER_S:
        conversion_factor = (
            1 / SEC_PER_YR / np.array(grid.cell_areas).reshape(grid.shape).T
        ) * 1e9
    else:
        raise NotImplementedError(f"Unknown {unit=}")

    unit_str = str(unit.value)
    if unit in PER_CELL_UNITS:
        unit_str += " cell-1"

    ds = xr.Dataset(
        data_vars=(
            {
                var_name_format.format(substance=sub, category=cat): (
                    [lat_name, lon_name],
                    inv.gdf[(cat, sub)].to_numpy().reshape(grid.shape).T
                    * conversion_factor,
                    {
                        "standard_name": f"{sub}_{cat}",
                        "long_name": f"{sub}_{cat}",
                        "units": unit_str,
                        "comment": f"emissions of {sub} in {cat}",
                        "projection": f"{crs}",
                    },
                )
                for sub in inv.substances
                for cat in inv.categories
                if (cat, sub) in inv.gdf
            }
            if not group_categories
            else {
                var_name_format.format(substance=sub): (
                    ["category", lat_name, lon_name],
                    np.array(
                        [
                            (
                                inv.gdf[(cat, sub)].to_numpy().reshape(grid.shape).T
                                if (cat, sub) in inv.gdf
                                else np.zeros(grid.shape).T
                            )
                            for cat in inv.categories
                        ]
                    )
                    * conversion_factor,
                    {
                        "standard_name": f"tendency_of_atmosphere_mass_content_of_{sub}_due_to_emission",
                        "long_name": f"{sub}",
                        "units": unit_str,
                        "comment": f"emissions of {sub}",
                        "projection": f"{crs}",
                    },
                )
                for sub in inv.substances
            }
        ),
        coords={
            "substance": inv.substances,
            "category": inv.categories,
            # Grid coordinates
            lon_name: (
                lon_name,
                grid.lon_range,
                {
                    "standard_name": "longitude",
                    "long_name": "longitude",
                    "units": "degrees_east",
                    "comment": "center_of_cell",
                    "bounds": f"{lon_name}_bnds",
                    "projection": f"{crs}",
                    "axis": "X",
                },
            ),
            lat_name: (
                lat_name,
                grid.lat_range,
                {
                    "standard_name": "latitude",
                    "long_name": "latitude",
                    "units": "degrees_north",
                    "comment": "center_of_cell",
                    "bounds": f"{lat_name}_bnds",
                    "projection": f"{crs}",
                    "axis": "Y",
                },
            ),
            f"{lon_name}_bnds": (
                f"{lon_name}_bnds",
                grid.lon_bounds,
                {
                    "standard_name": "longitude",
                    "long_name": "longitude of cell boundaries",
                    "units": "degrees_east",
                    "comment": "cell_boundaries",
                    "axis": "X",
                },
            ),
            f"{lat_name}_bnds": (
                f"{lat_name}_bnds",
                grid.lat_bounds,
                {
                    "standard_name": "latitude",
                    "long_name": "latitude of cell boundaries",
                    "units": "degrees_north",
                    "comment": "cell_boundaries",
                    "axis": "Y",
                },
            ),
        },
        attrs=netcdf_attributes,
    )

    if add_totals:
        for sub in inv.substances:
            if group_categories:
                da = ds[var_name_format.format(substance=sub)].sum("category")
            else:
                var_names = [
                    var_name_format.format(substance=sub, category=cat)
                    for cat in inv.categories
                    if var_name_format.format(substance=sub, category=cat)
                    in ds.data_vars
                ]
                da = sum([ds[name] for name in var_names])

            ds[f"emi_{sub}_all_sectors"] = da
            ds[f"emi_{sub}_all_sectors"].attrs = {
                "standard_name": f"tendency_of_atmosphere_mass_content_of_{sub}_due_to_emission",
                "long_name": f"Aggregated Emissions of {sub} from all sectors",
                "units": unit_str,
                "comment": "annual mean emission rate",
                "projection": f"{crs}",
            }
            if group_categories:
                total_emission = ds[var_name_format.format(substance=sub)]
            else:
                da_of_cat = {
                    cat: (
                        ds[var_name_format.format(substance=sub, category=cat)]
                        if var_name_format.format(substance=sub, category=cat)
                        in ds.data_vars
                        else xr.zeros_like(da)
                    )
                    for cat in inv.categories
                }
                total_emission = xr.concat(list(da_of_cat.values()), dim="category")
            # Total emission is not weighted by the cell area
            # So we always give kg/year
            if unit == Units.KG_PER_M2_PER_S:
                total_emission = (
                    total_emission
                    * np.array(grid.cell_areas).reshape(grid.shape).T
                    * SEC_PER_YR
                )
            elif unit == Units.KG_PER_YEAR:
                pass
            elif unit == Units.MUG_PER_M2_PER_S:
                total_emission = (
                    total_emission
                    * np.array(grid.cell_areas).reshape(grid.shape).T
                    * SEC_PER_YR
                    * 1e-9
                )
            else:
                raise NotImplementedError(f"Unknown {unit=}")
            ds[f"emi_{sub}_total"] = total_emission.sum([lon_name, lat_name])
            ds[f"emi_{sub}_total"].attrs = {
                "long_name": f"Total Emissions of {sub}",
                "units": "kg yr-1",
                "comment": "annual total emission",
            }

    # add the cell area
    ds["cell_area"] = (
        [lat_name, lon_name],
        np.array(grid.cell_areas).reshape(grid.shape).T,
        {
            "standard_name": "cell_area",
            "long_name": "cell_area",
            "units": "m2",
            "comment": "area of the cell",
            "projection": f"{crs}",
        },
    )
    path = Path(path)
    out_filepath = path.with_suffix(".nc")
    ds.to_netcdf(out_filepath)

    return out_filepath



# TODO: check if we need to use another source of data
# constants
country_codes = {
    "ALB": 1,
    "AUT": 2,
    "BEL": 3,
    "BGR": 4,
    "DNK": 6,
    "FIN": 7,
    "FRA": 8,
    "SMR": 8, #HJM
    "FGD": 9,
    "FFR": 10,
    "GRC": 11,
    "HUN": 12,
    "IRL": 14,
    "ITA": 15,
    "LUX": 16,
    "NLD": 17,
    "NOR": 18,
    "POL": 19,
    "PRT": 20,
    "ROM": 21,
    "ROU": 21,
    "ESP": 22,
    "AND": 22, # Andora -> ESP
    "SWE": 23,
    "CHE": 24,
    "TUR": 25,
    "GBR": 27,
    "IMN": 27, # Isle of Man -> GBR
    "GGY": 27, # Guernsey -> GBR
    "JEY": 27, # Jursey -> GBR
    "BLR": 39,
    "UKR": 40,
    "MKD": 41,
    "MDA": 42,
    "EST": 43,
    "LVA": 44,
    "LTU": 45,
    "CZE": 46,
    "SVK": 47,
    "SVN": 48,
    "HRV": 49,
    "BIH": 50,
    "YUG": 51,
    "GEO": 54,
    "MLT": 57,
    "DEU": 60,
    "RUS": 61,
    "ARM": 56,
    "AZE": 58,
    "CYP": 55,
    "ISL": 18,  # added for AQMEII, assumed to be same as Norway
    # ISO 2 (from EMEP site)
    "AL": 1,  # Albania
    "AT": 2,  # Austria
    "BE": 3,  # Belgium
    "BG": 4,  # Bulgaria
    "FCS": 5,  # Former Czechoslovakia
    "DK": 6,  # Denmark
    "FI": 7,  # Finland
    "ALD": 7,  # Aland, part of Finland
    "FR": 8,  # France
    "FGD": 9,  # Former German Democratic Republic
    "FFR": 10,  # Former Federal Republic of Germany
    "GR": 11,  # Greece
    "WSB": 11,  # Akrotiri/Santorini (Assigned to Greece)
    "HU": 12,  # Hungary
    "IS": 13,  # Iceland
    "IE": 14,  # Ireland
    "IT": 15,  # Italy
    "LU": 16,  # Luxembourg
    "NL": 17,  # Netherlands
    "NO": 18,  # Norway
    "PL": 19,  # Poland
    "PT": 20,  # Portugal
    "RO": 21,  # Romania
    "ES": 22,  # Spain
    "SMR": 22,  # San Marino (assigned to Spain)
    "AD": 22,  # Andorra (assigned to Spain)
    "GIB": 22,  # Gibraltar (assigned to Spain)
    "SE": 23,  # Sweden
    "CH": 24,  # Switzerland
    "TR": 25,  # Turkey
    "FSU": 26,  # Former USSR
    "GB": 27,  # United Kingdom
    "FRO": 27,  # Faroe Island (assigned to United Kingdom)
    "VOL": 28,  # Volcanic emissions
    "REM": 29,  # Remaining land Areas
    "BAS": 30,  # Baltic Sea
    "NOS": 31,  # North Sea
    "ATL": 32,  # Remaining North-East Atlantic Ocean
    "MED": 33,  # Mediterranean Sea
    "BLS": 34,  # Black Sea
    "NAT": 35,  # Natural marine emissions
    "RUO": 36,  # Kola & Karelia
    "RUP": 37,  # St.Petersburg & Novgorod-Pskov
    "RUA": 38,  # Kaliningrad
    "BY": 39,  # Belarus
    "UA": 40,  # Ukraine
    "MD": 41,  # Republic of Moldova
    "RUR": 42,  # Rest of the Russian Federation
    "EE": 43,  # Estonia
    "LV": 44,  # Latvia
    "LT": 45,  # Lithuania
    "CZ": 46,  # Czech Republic
    "SK": 47,  # Slovakia
    "SI": 48,  # Slovenia
    "HR": 49,  # Croatia
    "BA": 50,  # Bosnia and Herzegovina
    "CS": 51,  # Serbia and Montenegro
    "MK": 52,  # The former Yugoslav Republic of Macedonia
    "KZ": 53,  # Kazakhstan in the former official EMEP domain
    "GE": 54,  # Georgia
    "CY": 55,  # Cyprus
    "CYN": 55,  # Cyprus (alternate code)
    "CNM": 55,  # Cyprus (alternate code)
    "ESB": 55,  # Dhekelia Cantonment (assigned to Cyprus)
    "AM": 56,  # Armenia
    "MT": 57,  # Malta
    "ASI": 58,  # Remaining Asian areas
    "LI": 59,  # Liechtenstein,
    "LIE": 59,  # Liechtenstein (alternate code)
    "DE": 60,  # Germany
    "RU": 61,  # Russian Federation in the former official EMEP domain
    "MC": 62,  # Monaco
    "MCO": 62,  # Monaco
    "NOA": 63,  # North Africa
    "MAR": 63,  # Maroko (assigned to NOA)
    "TUN": 63,  # Tunisia (assigned to NOA)
    "DZA": 63,  # Algeria (assigned to NOA)
    "SYR": 63,  # Syria (assigned to NOA)
    "EU": 64,  # European Community
    "US": 65,  # United States
    "CA": 66,  # Canada
    "BIC": 67,  # Boundary and Initial Conditions
    "KG": 68,  # Kyrgyzstan
    "AZ": 69,  # Azerbaijan
    "ATX": 70,  # EMEP-external Remaining North-East Atlantic Ocean
    "RUX": 71,  # EMEP-external part of Russian Federation
    "RS": 72,  # Serbia
    "SRB": 72,  # Serbia (alternate code)
    "KOS": 72,  # Kosovo (assigned to SRB)
    "ME": 73,  # Montenegro
    "MNE": 73,  # Montenegro (alternate code)
    "RFE": 74,  # Rest of Russian Federation in the extended EMEP domain
    "KZE": 75,  # Rest of Kazakhstan in the extended EMEP domain
    "UZO": 76,  # Uzbekistan in the former official EMEP domain
    "TMO": 77,  # Turkmenistan in the former official EMEP domain
    "UZE": 78,  # Rest of Uzbekistan in the extended EMEP domain
    "TME": 79,  # Rest of Turkmenistan in the extended EMEP domain
    "CAS": 80,  # Caspian Sea
    "TJ": 81,  # Tajikistan
    "ARO": 82,  # Aral Lake in the former official EMEP domain
    "ARE": 83,  # Rest of Aral Lake in the extended EMEP domain
    "ASM": 84,  # Modified Remaining Asian Areas in the former official EMEP domain
    "ASE": 85,  # Remaining Asian Areas in the extended EMEP domain
    "AOE": 86,  # Arctic Ocean in the extended EMEP domain
    "RFX": 87,  # Extended EMEP External Part of Russian Federation
    "ASX": 88,  # Extended EMEP External Part of Asia
    "PAX": 89,  # Extended EMEP External Part of Pacific Ocean
    "AOX": 90,  # Extended EMEP External Part of Arctic Ocean
    "NAX": 91,  # Extended EMEP External Part of North Africa
    "KZT": 92,  # Kazakhstan
    "RUE": 93,  # Russian Federation in the extended EMEP domain (RU+RFE+RUX)
    "UZ": 94,  # Uzbekistan
    "TM": 95,  # Turkmenistan
    "AST": 96,  # Asian areas in the extended EMEP domain (ASM+ASE+ARO+ARE+CAS)
    "FYU": 99,  # Former Yugoslavia
    # Added for emiproc 
    "ISR": 200, # Israel
    "LBN": 201, # Lebanon
    "LBY": 202, # Libya
    "VAT": 203, # Vatican City
    "ETH": 204, # Ethiopia
    "SOM": 205, # Somalia
    "SOL": 206, # ???
    "UZB": 207, # Uzbekistan
    "KAZ": 208, # Kazakhstan
    "TKM": 209, # Turkmenistan
    "KGZ": 210, # Kyrgyzstan
    "TJK": 211, # Tajikistan
    "IRN": 212, # Iran
    "IRQ": 213, # Iraq
    # End of added for emiproc
    "BEF": 301,  # Belgium (Flanders)
    "BA2": 302,  # Baltic Sea EU Cargo o12m
    "BA3": 303,  # Baltic Sea ROW Cargo o12m
    "BA4": 304,  # Baltic Sea EU Cargo i12m
    "BA5": 305,  # Baltic Sea ROW Cargo i12m
    "BA6": 306,  # Baltic Sea EU Ferry o12m
    "BA7": 307,  # Baltic Sea ROW Ferry o12m
    "BA8": 308,  # Baltic Sea EU Ferry i12m
    "BA9": 309,  # Baltic Sea ROW Ferry i12m
    "NO2": 312,  # North Sea EU Cargo o12m
    "NO3": 313,  # North Sea ROW Cargo o12m
    "NO4": 314,  # North Sea EU Cargo i12m
    "NO5": 315,  # North Sea ROW Cargo i12m
    "NO6": 316,  # North Sea EU Ferry o12m
    "NO7": 317,  # North Sea ROW Ferry o12m
    "NO8": 318,  # North Sea EU Ferry i12m
    "NO9": 319,  # North Sea ROW Ferry i12m
    "AT2": 322,  # Remaining North-East Atlantic Ocean EU Cargo     'o1': 2 # m
    "AT3": 323,  # Remaining North-East Atlantic Ocean ROW Cargo     'o1': 2 # m
    "AT4": 324,  # Remaining North-East Atlantic Ocean EU Cargo     'i1': 2 # m
    "AT5": 325,  # Remaining North-East Atlantic Ocean ROW Cargo     'i1': 2 # m
    "AT6": 326,  # Remaining North-East Atlantic Ocean EU Ferry     'o1': 2 # m
    "AT7": 327,  # Remaining North-East Atlantic Ocean ROW Ferry     'o1': 2 # m
    "AT8": 328,  # Remaining North-East Atlantic Ocean EU Ferry     'i1': 2 # m
    "AT9": 329,  # Remaining North-East Atlantic Ocean ROW Ferry     'i1': 2 # m
    "ME2": 332,  # Mediterranean Sea EU Cargo o12m
    "ME3": 333,  # Mediterranean Sea ROW Cargo o12m
    "ME4": 334,  # Mediterranean Sea EU Cargo i12m
    "ME5": 335,  # Mediterranean Sea ROW Cargo i12m
    "ME6": 336,  # Mediterranean Sea EU Ferry o12m
    "ME7": 337,  # Mediterranean Sea ROW Ferry o12m
    "ME8": 338,  # Mediterranean Sea EU Ferry i12m
    "ME9": 339,  # Mediterranean Sea ROW Ferry i12m
    "BL2": 342,  # Black Sea EU Cargo o12m
    "BL3": 343,  # Black Sea ROW Cargo o12m
    "BL4": 344,  # Black Sea EU Cargo i12m
    "BL5": 345,  # Black Sea ROW Cargo i12m
    "BL6": 346,  # Black Sea EU Ferry o12m
    "BL7": 347,  # Black Sea ROW Ferry o12m
    "BL8": 348,  # Black Sea EU Ferry i12m
    "BL9": 349,  # Black Sea ROW Ferry i12m
    "GL": 601,  # Greenland
}

# Mapping from code to iso3
code_2_iso3 = {
    code: iso3 for iso3, code in country_codes.items() if len(iso3) == 3
}



"""Heating degree days (HDD) profile."""
import numpy as np
import pandas as pd
from emiproc.profiles.temporal_profiles import TemporalProfile, create_scaling_factors_time_serie

def create_HDD_scaling_factor(
    serie_T: pd.Series,
    heating_profile: list[TemporalProfile],
    dhw_profile: list[TemporalProfile],
    min_heating_T: float = 12.0,
    inside_T: float = 20.0,
    dhw_scaling: float = 5.59 / 24,
) -> pd.Series:
    """Generate the scaling factor for the heating degree days formula.

    The HDD formula procceds this way: 
    first Calculate the mean temprature of the day, and the heating demeand for the day 

    .. math::
        HDD = (T_{inside} - T_{mean}) 
    
    Where :math:`T_{inside}` is the inside temperature
    and :math:`T_{mean}` is the mean temperature of the day.

    If :math:`T_{mean} < T_{min}` then the heating is not activated and
    :math:`HDD = 0`.

    A day of year profile can then be calculated using 

    .. math::
        a_{HDD} = \\frac{HDD}{\\overline{HDD}}
    
    Where :math:`\\overline{HDD}` is the yearly average of the HDD.

    A profile for domestic hot water is then added to this profile. 

    .. math::
        a_{H} = (1 - f_{DHW}) * HDD_{H} * a_{HDD} + a_{DHW} * f_{DHW}

    Where :math:`HDD_{H}` is the hourly heating profile and :math:`a_{DHW}` is the
    hourly domestic hot water profile and :math:`f_{DHW}` is the scaling factor for
    the domestic hot water profile.

    Thus the return profile correspond to both heating and domestic hot water,
    with a hourly resolution.

    :arg serie_T: the timeserie of the temperature (in Celsius)
    :arg heating_profile: the heating profile
    :arg dhw_profile: the domestic hot water profile
    :arg min_heating_T: the minimum temperature for which heating is activated
    :arg inside_T: the inside temperature
    :arg dhw_scaling: :math:`f_{DHW}`, the scaling of domesting hot water demand vs surface heating
        must be a ratio between 0 and 1. If 0, the profiles will correspond to 
        space heating only.
    """
    serie_daily_T = serie_T.resample("D").mean()

    # Heating degrees of the day
    heating_activated = serie_daily_T < min_heating_T
    HDD = (inside_T - serie_daily_T) * heating_activated.astype(float)

    # Yearly avergage
    yearly_means = HDD.resample("Y").mean()
    ts_mean = pd.Series(np.nan, index=HDD.index)
    for dt, mean in yearly_means.items():
        ts_mean.loc[ts_mean.index.year == dt.year] = mean
    # Scale with the yearly means
    a_HDD = HDD / ts_mean

    start = serie_T.index.min()
    end = serie_T.index.max()
    hdd_ts = pd.Series(np.nan, pd.date_range(start, end, freq="H"))
    # Get hourly values
    a_HDD_hourly = (
        a_HDD.reindex(index=a_HDD.index.union(hdd_ts.index))
        .interpolate("pad")
        .reindex(hdd_ts.index)
    )

    heating_ts = create_scaling_factors_time_serie(start, end, heating_profile)
    dhw_ts = create_scaling_factors_time_serie(start, end, dhw_profile)

    return (1. - dhw_scaling) * a_HDD_hourly * heating_ts + dhw_ts * dhw_scaling




from __future__ import annotations

import logging
from pathlib import Path
from dataclasses import dataclass
from enum import Enum, auto
from os import PathLike
import numpy as np
import pandas as pd
import xarray as xr

from emiproc import deprecated
from emiproc.profiles.utils import (
    get_profiles_indexes,
    merge_indexes,
    read_profile_csv,
    read_profile_file,
)
from emiproc.profiles import naming

logger = logging.getLogger(__name__)


@dataclass
class VerticalProfile:
    """Vertical profile.

    A vertical profile defines how the emission is split vertically on the
    altitude. A vertical profile is defined simply by its ratios
    and the height levels.

    You can check the conditions required on the profile in
    :py:func:`check_valid_vertical_profile`

    :arg ratios: The proportion of emission that is in each layer.
    :arg height: The top height of the layers.
        The first layer starts at 0 meter and ends at height[0].
        The second layer starts at height[0] and ends at height[1].
        Over the last height value, there is no emission.
    """

    ratios: np.ndarray
    height: np.ndarray

    @property
    def n_profiles(self) -> int:
        return 1


@dataclass
class VerticalProfiles:
    """Vertical profiles.

    This is very similar to :py:class:`VerticalProfile` but it can store
    many ratios for the same height distributions

    :arg ratios: a (n_profiles, n_heights) array.
    :arg height: Same as :py:attr:`VerticalProfile.height` .
    """

    ratios: np.ndarray
    height: np.ndarray

    def __post_init__(self):
        # Make sure numpy arrays
        self.ratios = np.asarray(self.ratios)
        self.height = np.asarray(self.height)

        assert self.ratios.shape[1] == len(
            self.height
        ), "Ratios and height must have the same length"

    @property
    def n_profiles(self) -> int:
        return self.ratios.shape[0]

    @property
    def n_heights(self) -> int:
        return self.ratios.shape[1]

    def copy(self):
        """Make a deep copy of the profiles."""
        return VerticalProfiles(
            self.ratios.copy(),
            self.height.copy(),
        )

    def __add__(self, other: VerticalProfiles):
        if isinstance(other, int) and other == 0:
            # Useful for call in sum()
            return self.copy()

        assert isinstance(other, VerticalProfiles)
        assert np.allclose(other.height, self.height)

        return VerticalProfiles(
            height=self.height,
            ratios=np.concatenate([self.ratios, other.ratios], axis=0),
        )

    def __radd__(self, other: VerticalProfiles):
        # Useful for call in sum()
        return self.__add__(other)

    def __getitem__(self, index: int) -> VerticalProfile:
        return VerticalProfile(
            ratios=self.ratios[index],
            height=self.height,
        )

    def __len__(self) -> int:
        return self.n_profiles

    @property
    def scaling_factors(self) -> np.ndarray:
        """Get the scaling factors for the profiles."""
        return self.ratios * self.n_heights


class GroupingMethod(Enum):
    KEEP_ALL_LEVELS = auto()


def get_mid_heights(max_heights: np.ndarray) -> np.ndarray:
    """Get the mid position of the height based on emiproc convention for height levels."""
    min_height = np.roll(max_heights, 1)
    # First level start at 0
    min_height[0] = 0
    mid_heights = (max_heights + min_height) / 2
    return mid_heights


def get_delta_h(max_heights: np.ndarray) -> np.ndarray:
    """Get height coverd by each level based on emiproc convention for height levels.."""
    min_height = np.roll(max_heights, 1)
    # First level start at 0
    min_height[0] = 0
    return max_heights - min_height


def get_weights_profiles_interpolation(
    from_p: np.ndarray, to_p: np.ndarray
) -> np.ndarray:
    """Calculate the weights matrix between two profiles.

    The weights matrix can then simply be used .

    The two given profiles must be sorted.
    It is assumed that they are vertical profiles from emiproc convention
    """

    # Initialize parameters for the algorithm
    i, j = 0, 0
    last = 0.0

    diff = np.zeros((len(to_p), len(from_p)))

    while i < len(from_p) and j < len(to_p):
        # Will check the distance with the last point
        # and will input it in the diff matrix
        f = from_p[i]
        t = to_p[j]
        if f <= t:
            diff[j, i] = f - last
            i += 1
            last = f
        else:
            diff[j, i] = t - last
            j += 1
            last = t

    if j == len(to_p):
        # Assing the last ones to the last line
        while i < len(from_p):
            f = from_p[i]
            diff[-1, i] += f - last
            i += 1
            last = f

    # Weights is the noramlized
    return diff / diff.sum(axis=0)


def resample_vertical_profiles(
    *profiles: VerticalProfile | VerticalProfiles,
    specified_levels: np.ndarray | None = None,
) -> VerticalProfiles:
    """Resample vertical profiles into one vertical profiles object.

    Allows for profiles of different height levels to be groupped into one.
    Sample the profile on the heights level given.

    Uses a conservative interpolation method, that ensure that
    even on higher resolution the profile will be exactly the same.
    Note that this sometimes has no physical sense and a linear
    interpolation when using profiles would be better
    at higher resolutions, but this has to be a choice from the user.

    :arg specified_levels: If this is specified, you can select an arbitray
        scale on which to reproject. If not specified, all the levels found
        in the profiles will be used.
        The ordering of the profiles will match the order given as input.
    """

    # Find the levels we want to use
    if specified_levels is None:
        levels = np.unique(np.concatenate([p.height for p in profiles]))
    else:
        levels = specified_levels

    out_ratios = []
    for p in profiles:
        # Get the weights for remapping those profiles
        weights = get_weights_profiles_interpolation(p.height, levels)

        # Do the remapping and add it to the results
        out_ratios.append(p.ratios.dot(weights.T))

    return VerticalProfiles(np.vstack(out_ratios), levels)


def check_valid_vertical_profile(vertical_profile: VerticalProfile | VerticalProfiles):
    """Check that the vertical profile meets requirements.

    * height must have positive values
    * height must have strictly increasing values
    * ratios must sum up to one
    * ratios must all be >= 0
    * ratios and height must have the same len
    * no nan values in any of the arrays

    :arg veritcal_profile: The profile to check.
    :raises AssertionError: If the profile is invalid.
    """

    assert isinstance(vertical_profile, (VerticalProfile, VerticalProfiles))

    h = vertical_profile.height
    r = vertical_profile.ratios

    assert np.all(~np.isnan(r)) and np.all(~np.isnan(h)), "Cannot contain nan values"

    assert np.all(h > 0)
    assert np.all(h[1:] > np.roll(h, 1)[1:]), f"height must be increasing, got {h=}"
    if isinstance(vertical_profile, VerticalProfile):
        assert np.sum(r) == 1.0
        assert len(r) == len(h)
    else:
        ratio_sum = np.sum(r, axis=1)
        assert np.allclose(ratio_sum, 1.0), f"Ratios must sum to 1, but {ratio_sum=}"
        assert r.shape[1] == len(h)
    assert np.all(r >= 0)


@deprecated
def from_csv(file: PathLike) -> tuple[VerticalProfiles, list[str | tuple[str, str]]]:
    """Read a csv file containing vertical profiles.

    ..warning:: This is deprecated. Use :py:func:`read_vertical_profiles` instead.


    The format is the following::

        Category,Substance,20m,92m,184m,324m,522m,781m,1106m
        Public_Power,CO2,0,0,0.0025,0.51,0.453,0.0325,0.002
        Public_Power,CH4,0,0,0.0025,0.51,0.453,0.0325,0.002
        Industry,CO2,0.06,0.16,0.75,0.03,0,0,0
        ...

    The first line is the header, the first column is the category,
    the second column is the substance and the rest of the columns
    are the ratios for each height level.

    The heights level specify mean from the previous height to this one.
    (e.g. 20m is the mean from 0 to 20m, 92m is the mean from 20m to 92m, ...)

    The Substance column is optional, if not present, the category will be used
    only.

    """

    # Read the file
    df, cat_header, sub_header = read_profile_csv(file)

    # Get the height levels
    heights_mapping = {
        float(col_name[:-1]): col_name
        for col_name in df.columns
        # Take only the columns that end with m
        if col_name.endswith("m")
    }
    heights = sorted(heights_mapping.keys())
    heights_cols = [heights_mapping[h] for h in heights]

    # Create the profiles
    profiles = VerticalProfiles(df[heights_cols].to_numpy(), heights)

    if sub_header is None:
        cat_sub = df[cat_header].to_list()
    else:
        cat_sub = df[[cat_header, sub_header]].apply(tuple, axis=1).to_list()

    return profiles, cat_sub


def read_vertical_profiles(
    profiles_dir: PathLike,
) -> tuple[VerticalProfiles, xr.DataArray]:
    """Read vertical profiles from csv files.

    Vertical profiles only depend on the category.

    The format expected from the files is:
    Line starting with # are ignored.
    A header line should contain a column for the category, substance, country ...

    The other columns should contain the value of the levels.
    The columns must be sorted from smallest to largest.
    They should be called one of the following:
    - using the ending height: ex.  12;23;76
    - using the interval ex. 12-23;23-76


    You can add a m  (20m, 40m) to specify it is meters, but it is not required.
    The height is always assumed to be of meter units.

    An example of a file is::

        # This is a comment
        # Any number of comment lines are allowed
        # Below is the header line
        Category,Substance,20m,92m,184m,324m,522m,781m,1106m
        Public_Power,CO2,0,0,0.0025,0.51,0.453,0.0325,0.002
        Public_Power,CH4,0,0,0.0025,0.51,0.453,0.0325,0.002
        Industry,CO2,0.06,0.16,0.75,0.03,0,0,0
        ...


    :arg profiles_dir: The directory containing the profiles.
        The profiles file must contain the word "height" in their name.
        Alternatively you can provide a path to a specific file.
        For compatiblity with tno, you can also include the words "area" or "point"
        in the name of the file to specify if the profiles are for gridded emissions
        or for shapped emissions.

    :return: A tuple containing the vertical profiles and an xarray mapping
        which kind of emission correspond to which profile.
    """
    profiles_dir = Path(profiles_dir)

    if profiles_dir.is_file():
        v_profiles_files = [profiles_dir]
    elif profiles_dir.is_dir():
        v_profiles_files = [f for f in profiles_dir.glob("*.csv") if "height" in f.stem]
    else:
        raise FileNotFoundError(
            f"Vertical profiles directory {profiles_dir} is not a directory."
        )

    # Check if we found any file
    if len(v_profiles_files) == 0:
        logger.warning(f"No vertical profiles found in {profiles_dir}.")
        return None, None

    # These are the names in the sectors column of tno
    # seems that these should be changed with new versions

    v_profiles = []
    v_profiles_indexes = []
    index_offset = 0

    for file in v_profiles_files:
        logger.debug(f"Reading vertical profiles from '{file}' .")
        df_vertical = read_profile_file(file)

        boundarys = [
            col
            for col in df_vertical.columns
            if col.strip() not in naming.all_reserved_colnames
        ]
        logger.debug(f"Vertical boundaries columns found: {boundarys}")
        starts = []
        ends = []
        end = 0
        try:
            for boundary_str in boundarys:
                for char in ["m", " ", "\t"]:
                    boundary_str = boundary_str.replace(char, "")

                if "-" in boundary_str:
                    # Specified interval
                    a, b = boundary_str.split("-")
                    start = float(a)
                    end = float(b)
                else:
                    # Specified only the end of the interval
                    start = end  # Use the previous end
                    end = float(boundary_str)
                # Save
                starts.append(start)
                ends.append(end)
        except Exception as e:
            raise ValueError(
                f"Cound not interpret the vertical boundaries in '{file}' . Read"
                f" {boundarys=}.\n Please check the format of the file. For more"
                " information, see:"
                " https://emiproc.rtfd.io/en/latest/api.html#emiproc.profiles.vertical_profiles.read_vertical_profiles "
            ) from e
        tops = np.array(ends)
        bots = np.array(starts)
        if bots[0] != 0:
            raise ValueError(
                f"The first vertical boundary should be 0, not {bots[0]}.\n Check the"
                f" file {file}."
            )

        # Store the profiles in the object and check the validity
        profile = VerticalProfiles(df_vertical[boundarys].to_numpy(), tops)
        try:
            check_valid_vertical_profile(profile)
        except AssertionError as e:
            raise ValueError(
                f"Vertical profile in {file} is not valid. Please check the format of"
                " the file."
            ) from e
        v_profiles.append(profile)

        # all the profiles (time and vertical)
        indexes = get_profiles_indexes(df_vertical)
        # Add the offset where the indexes is defined (!= -1)
        indexes = indexes.where(indexes == -1, indexes + index_offset)

        if not "type" in indexes.dims:
            if "area" in file.stem:
                indexes = indexes.expand_dims({"type": ["gridded"]})
            elif "point" in file.stem:
                indexes = indexes.expand_dims({"type": ["shapped"]})
            else:
                logger.debug(
                    f"Could not determine a type of profile in {file.stem}. Assuming"
                    " applies to all types."
                )

        v_profiles_indexes.append(indexes)
        index_offset += len(profile)

    logger.debug(f"combining {v_profiles_indexes=}")
    combined_indexes = merge_indexes(v_profiles_indexes)

    return resample_vertical_profiles(*v_profiles), combined_indexes



import logging
from os import PathLike
from pathlib import Path
from urllib.error import HTTPError
import xarray as xr
import numpy as np
import re
import geopandas as gpd
import pyogrio

import urllib.request
import zipfile

from emiproc.grids import WGS84, EDGARGrid, RegularGrid
from emiproc.inventories import Inventory
from emiproc.utilities import SEC_PER_YR


def download_edgar_files(
    data_dir: PathLike,
    year: int,
    categories: list[str] = [
        "ENE",
        "REF_TRF",
        "IND",
        "TNR_Aviation_CDS",
        "TNR_Aviation_CRS",
        "TNR_Aviation_LTO",
        "TNR_Aviation_SPS",
        "TRO",
        "TNR_Other",
        "TNR_Ship",
        "RCO",
        "PRO_FFF",
        "NMM",  # Non-metallic minerals production
        "CHE",  # Chemical processes
        "IRO",  # Iron and steel production
        "NFE",  # Non-ferrous metals production
        "NEU",  # Non-energy use of fuels
        "PRU_SOL",  # Solvents and products use
        # Agriculture
        "ENF",  # Enteric fermentation
        "MNM",  # Manure management
        "AWB",  # Agricultural waste burning
        "AGS",  # Agricultural soils
        "N2O",  # Indirect N2O emissions from agriculture
        "SWD_LDF",  # Solid waste landfills
        "SWD_INC",  # Solid waste incineration
        "WWT",  # Waste water handling
        "IDE",  # Indirect emissions from NOx and NH3
    ],
    substances: list[str] = [
        "CO2",
        "CH4",
        "N2O",
        "CO2bio",
        "GWP_100_AR5_GHG",
    ],
    link_template: str = "https://jeodpp.jrc.ec.europa.eu/ftp/jrc-opendata/EDGAR/datasets/v80_FT2022_GHG/{substance}/{category}/emi_nc/v8.0_FT2022_GHG_{substance}_{year}_{category}_emi_nc.zip",
):
    download_links = [
        link_template.format(substance=substance, category=category, year=year)
        for substance in substances
        for category in categories
    ]

    downloaded = []

    # Download the files
    for link in download_links:
        filename = link.split("/")[-1]
        filepath = data_dir / filename
        try:
            urllib.request.urlretrieve(link, filepath)
        except HTTPError as e:
            # Link does not exist
            continue

        with zipfile.ZipFile(filepath, "r") as zip_ref:
            zip_ref.extractall(data_dir)

        downloaded.append(filepath)
        filepath.unlink()

    if not downloaded:
        raise ValueError(
            f"No files were downloaded for {substances=} {year=} {categories=}."
            "Check the links."
        )

    print(f"Downloaded {len(downloaded)} files.")


class EDGARv8(Inventory):
    """The EDGAR inventory.

    `Emissions Database for Global Atmospheric Research <https://edgar.jrc.ec.europa.eu/>`_


    The files are freely available.
    There are different versions. The one we use here is the
    `Annual sector-specific gridmaps` and we scale them over the months using the
    `Monthly time series` files.

    Download the files you want to include and provide the path to the directory
    containing the files.

    EDGAR has only grid cell sources.
    """

    grid: EDGARGrid

    def __init__(
        self,
        nc_file_pattern_or_dir: PathLike,
        year: int | None = None,
        use_short_category_names: bool = False,
    ) -> None:
        """Create a EDGAR_Inventory.

        :arg nc_file_pattern_or_dir: Pattern or directory of files.
        :arg year: Year of the inventory.
        :arg use_short_category_names: Use short category names.

        """
        super().__init__()

        nc_file_pattern = Path(nc_file_pattern_or_dir)
        logger = logging.getLogger(__name__)

        if nc_file_pattern.is_dir():
            nc_file_pattern = nc_file_pattern / "*.nc"

        columns = {}

        file_for_grid = None

        for filepath in nc_file_pattern.parent.glob(nc_file_pattern.name):
            with xr.open_dataset(filepath) as ds:
                if "emissions" not in ds:
                    logger.warning(
                        f"File {filepath} does not contain 'emissions' variable. Skipping."
                    )
                    continue
                da = ds["emissions"]
                substance = da.attrs["substance"]
                category = da.attrs["long_name"]
                if use_short_category_names:
                    # Read from the path string
                    category = "_".join(filepath.stem.split("_")[5:-1])

                file_year = int(da.attrs["year"])
                if year is None:
                    year = file_year
                if file_year != year:
                    logger.warning(
                        f"File {filepath} has year {file_year}, expected {year}. Skipping."
                    )
                    continue
                units = da.attrs["units"]
                assert units == "Tonnes", f"Units are {units}, expected `Tonnes`."

                columns[(category, substance)] = da.data.T.reshape(-1)

            if file_for_grid is None:
                file_for_grid = filepath

        with xr.open_dataset(file_for_grid) as ds:
            self.grid = RegularGrid.from_centers(
                ds["lon"].data, ds["lat"].data, name="EDGARv8_grid"
            )

        # Swtich longitude from 0/360 to -180/180
        self.gdf = gpd.GeoDataFrame(
            data={
                # Convert from tonnes/yr to kg/yr
                col: value * 1e3
                for col, value in columns.items()
            },
            geometry=self.grid.gdf.geometry,
            crs=WGS84,
        )

        # Empty shaped sources
        self.gdfs = {}


class EDGAR_Inventory(Inventory):
    """The EDGAR inventory.

    `Emissions Database for Global Atmospheric Research <https://edgar.jrc.ec.europa.eu/>`_

    EDGAR has only grid cell sources.
    """

    grid: EDGARGrid

    def __init__(
        self, nc_file_pattern: PathLike, grid_shapefile: PathLike | None = None
    ) -> None:
        """Create a EDGAR_Inventory.

        The EDGAR directory that contains the original datasets must be structured as ./substance/categories/dataset,
        e.g., ./SF6/NFE/v7.0_FT2021_SF6_*_NFE.0.1x0.1.nc and ./SF6/PRU/v7.0_FT2021_SF6_*_PRU.0.1x0.1.nc

        :arg nc_file_pattern: Pattern of files, e.g "EDGAR/SF6/PRU/v7.0_FT2021_SF6_*_PRU.0.1x0.1.nc"

        """
        super().__init__()

        nc_file_pattern = Path(nc_file_pattern)

        self.name = nc_file_pattern.stem

        list_filepaths = list(nc_file_pattern.parent.glob(nc_file_pattern.name))
        list_ds = [xr.open_dataset(f) for f in list_filepaths]

        substance = list_filepaths[0].parent.parent.stem

        substances_mapping = {}
        categories = [
            re.search(r"{}_(.+?)_(.+?)\.".format(substance), filepath.name)[2]
            for filepath in list_filepaths
        ]

        for i, ds in enumerate(list_ds):
            old_varname = list(ds.data_vars.keys())[0]
            new_varname = substance.upper()
            list_ds[i] = ds.rename_vars({old_varname: new_varname})
            substances_mapping[new_varname] = new_varname.upper()

        ds = xr.merge(list_ds)

        # Swtich longitude from 0/360 to -180/180
        attributes = ds.lon.attrs
        ds = ds.assign(lon=(ds.lon + 180) % 360 - 180).sortby("lon")
        ds.lon.attrs = attributes

        self.grid = EDGARGrid(list_filepaths[0])

        if grid_shapefile is None:
            polys = self.grid.cells_as_polylist
        else:
            gdf_geometry = gpd.read_file(grid_shapefile, engine="pyogrio")
            gdf_geometry = gdf_geometry.to_crs(WGS84)
            polys = gdf_geometry.geometry.values

        # Index in the polygon list (from the gdf) (index start at 1)
        poly_ind = np.arange(self.grid.nx * self.grid.ny)

        self.gdfs = {}
        mapping = {}
        for cat_idx, cat_name in enumerate(categories):

            for sub_in_nc, sub_emiproc in substances_mapping.items():
                tuple_idx = (cat_name, sub_emiproc)
                if tuple_idx not in mapping:
                    mapping[tuple_idx] = np.zeros(len(polys))

                # Add all the area sources corresponding to that category
                np.add.at(mapping[tuple_idx], poly_ind, ds[sub_in_nc].data.T.flatten())

        self.gdf = gpd.GeoDataFrame(
            mapping,
            geometry=polys,
            crs=WGS84,
        )

        self.cell_areas = self.grid.cell_areas

        # -- Convert to kg/yr
        self.gdf[list(mapping)] *= SEC_PER_YR * self.cell_areas[:, np.newaxis]



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

from datetime import datetime

project = "emiproc"
copyright = f"2022-{datetime.now().year}, Empa"
author = "Lionel Constantin, Empa"
release = datetime.now().strftime("%Y.%m.%d")

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "sphinx.ext.autodoc",
    "enum_tools.autoenum",
    "nbsphinx",
]

templates_path = ["_templates"]
exclude_patterns = []

# autodoc_typehints = 'description'  # show type hints in doc body instead of signature

# Don't show class signature with the class' name.
autodoc_class_signature = "separated"

python_maximum_signature_line_length = 88

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "furo"
html_static_path = ["_static"]



import numpy as np
import xarray as xr

import pytest

from emiproc.utilities import get_country_mask
from emiproc.grids import RegularGrid

from emiproc.tests_utils.test_grids import regular_grid_africa

from emiproc import FILES_DIR


def test_create_simple_mask():
    arr = get_country_mask(
        regular_grid_africa,
        resolution="110m",
    )
    # check that there are some countries in there
    # Not just -1 values
    assert len(np.unique(arr)) > 1
    # Make sure GMB was not found
    assert "GMB" not in np.unique(arr)
    # MAke sure Mauritania was found
    assert "MRT" in np.unique(arr)


def test_with_fractions():
    da = get_country_mask(
        regular_grid_africa,
        return_fractions=True,
    )
    total_fractions = da.sum(dim="country")
    # Check that sum of fractions is never more than 1
    assert (
        (total_fractions.values <= 1) | (np.isclose(total_fractions.values, 1.0))
    ).all()
    assert (total_fractions.values >= 0).all()

    # Full in MRT
    assert da.sel(cell=78, country="MRT").values == 1.0
    # THis is shared between SEN and ocean
    assert da.sel(cell=26, country="SEN").values > 0.01
    assert da.sel(cell=26, country="SEN").values < 0.5
    assert total_fractions.sel(cell=26).values < 0.5
    # This is just ocean
    assert (total_fractions.sel(cell=0).values == 0).all()


@pytest.mark.parametrize(
    "expected_weight_file, kwargs",
    [
        (FILES_DIR / "test" / "test_country_mask.npy", {}),
        (FILES_DIR / "test" / "test_country_mask.nc", {"return_fractions": True}),
    ],
)
def test_save_mask(expected_weight_file, kwargs):
    weigth_file = expected_weight_file
    if weigth_file.exists():
        # Clean the test
        weigth_file.unlink()
    out1 = get_country_mask(
        regular_grid_africa,
        weight_filepath=weigth_file,
        **kwargs,
    )
    assert weigth_file.exists()
    # Call again to make sure it is loaded
    out2 = get_country_mask(
        regular_grid_africa,
        weight_filepath=weigth_file,
        **kwargs,
    )

    # Check that the two are the same
    if isinstance(out1, xr.DataArray):
        xr.testing.assert_equal(out1, out2)
    elif isinstance(out1, np.ndarray):
        np.testing.assert_array_equal(out1, out2)
    else:
        raise TypeError(f"Unexpected type {type(out1)}")


if __name__ == "__main__":
    pytest.main([__file__])



import pytest 
from emiproc.inventories import  EmissionInfo
from emiproc.tests_utils.test_inventories import inv




def test_emission_info():

    info = EmissionInfo()

    info2 = EmissionInfo(height=1.0, height_over_buildings=False)

def test_inventory_with_emission_info():

    info = EmissionInfo()

    def set_missing_info():
        # 2 Catergories are missing
        inv.emission_infos = {'adf': info}

    def set_correct_info():
        inv.emission_infos = {'adf': info, 'liku': info, 'test': info }
        
        assert inv.emission_infos['adf'].height == info.height

    # If a category is missing, an error should be raised
    pytest.raises(ValueError, set_missing_info )

    set_correct_info()


   




from shapely.geometry import Polygon
from emiproc.regrid import get_weights_mapping

from emiproc.tests_utils.test_inventories import inv, inv_with_pnt_sources
from emiproc.tests_utils import WEIGHTS_DIR

polys = [Polygon([(0, 0), (1, 0), (1, 1), (0, 1), (0, 0)])]
def test_get_weights_mapping():
    get_weights_mapping(
        WEIGHTS_DIR/ 'test_get_weights_mapping', inv.gdf, polys
        )

def test_no_weights():
    get_weights_mapping(
        None, inv.gdf, polys
        )